
// --------------------------------------------------------------------------------

void AvgAccelAll() {
  // average the raw accelerometer values over a given number of values
  AcXSumL += AcX; AcYSumL += AcY; AcZSumL += AcZ;
  if (AcAvgDiv < AcAvgDivMax) {
    // increase the scope of the averaging factor if necessary
    // note at 20ms sample, AvgDiv = 50 == 1 second rolling average
    AcAvgDiv++;
  } else {
    // once the accumulator has the correct number of samples roll out the average
    AcXSumL -= AcXAvgL; AcYSumL -= AcYAvgL; AcZSumL -= AcZAvgL;
  }
  AcXAvgL = AcXSumL / AcAvgDiv;
  AcYAvgL = AcYSumL / AcAvgDiv;
  AcZAvgL = AcZSumL / AcAvgDiv;
}

// ---------------------------------------------------------------------------------

void AvgGyrAll() {
  // called as a 20ms subtask, keeps a rolling average of the gyroscopes
  GyXSumL += GyX; GyYSumL += GyY; GyZSumL += GyZ;
  if (GyAvgDiv < GyAvgDivMax) {
    // increase the scope of the averaging factor if necessary
    // note at 20ms sample, AvgDiv = 50 == 1 second rolling average
    GyAvgDiv++;
  } else {
    // once the accumulator has the correct number of samples roll out the average
    GyXSumL -= GyXAvgL; GyYSumL -= GyYAvgL; GyZSumL -= GyZAvgL;
  }

  GyXAvgL = GyXSumL/GyAvgDiv;
  GyYAvgL = GyYSumL/GyAvgDiv;
  GyZAvgL = GyZSumL/GyAvgDiv;
}

// --------------------------------------------------------------------------------

void BackAway() {
  // Perform a series of tasks associated with 'Back Away' response.
  // These tasks are only performed in SafeMode 4, every 20ms.
  if (!RangeEn) {return;} // LTOF must be enabled

  if (BackAwayTask > 0) {DispMode = 9;}
  switch (BackAwayTask) {
    case 0: // Do nothing state
      BackAwayCnt = 0; return;
    case 1: // Initialise
      Lft_Step_Cnt = 0; Rht_Step_Cnt = 0;
      BackAwayTime = 0;
      // Ensure that there is nothing in front of the robot at the outset.
      // Run a 2 second timer to ensure all is clear.
      if (Range > 200) {BackAwayCnt ++; if (BackAwayCnt >= 100) {BackAwayTask++;}
      } else {BackAwayCnt = 0;} 
      break;
    case 2: // Look for an approaching target.
      // The code sits at task 2 until something approaches.
      BackAwayCnt = 0; Lft_Step_Cnt = 0; Rht_Step_Cnt = 0;
      if (Range < 150) {
        // Target within range, so back away.
        Set_Speeds(4);  // max speed for retreat
        RxByte = 0b00001000; BackAwayTask++;
      } break;
    case 3: // Track approaching target
      if (Range > 200) {
        BackAwayCnt++;    // Sustain transient trigger for 100ms}
        if  (BackAwayCnt >= 5) {RxByte = 0b00000000; BackAwayTime = 0; BackAwayTask++;}
      } else {BackAwayCnt = 0;}
      break;
    case 4: // Wait after target is no longer there
           if (Range > 200) {BackAwayTime++;}  // Advance the timer
      else if (Range < 150) {
        // Target has reappeared within the timer window, so back away again
        RxByte = 0b00001000; BackAwayTask = 3;
        BackAwayCnt = 0; return;
      }
      // Drive forward after 5 second wait
      if (BackAwayTime >= 250) {
        Set_Speeds(1);    // min speed for return
        BackAwayTime = 0; // zero timer
        RxByte = 0b00000100; BackAwayTask++;
      } break;
    case 5: // Return to start point using step counts
      // If target re-appears then go into reverse again
      if (Range < 150) {
        // Target has reappeared within the timer window, so back away again
        RxByte = 0b00001000; BackAwayTask = 3;
        BackAwayCnt = 0; return;
      }
      // Otherwise work on reducing step counts
      if (((Lft_Step_Cnt + Rht_Step_Cnt)/2) >= 0) {
        // Step counts have returned to zero so reset tasks
        RxByte = 0b00000000; BackAwayTask= 1;
      } break;
  }
}

// --------------------------------------------------------------------------------

void battRead() {
  // read the analogue voltage connected to the battery every 40ms
  // this frunction is no longer called due to the limitations of ADC2 under Wi-Fi
  BatVol = analogRead(BatPin);
  BatSum = BatSum + BatVol - BatAvg;
  BatAvg = BatSum/50;             // ADC is averaged over 50 readings (2 sec) to remove noise

  // force BattAvg high for debug. This line is normally commented out
  // BatAvg = BatHigh;

  // if (TEST) {Serial.println("Battery " + String(BatVol) + "\t" + String(float(BatAvg)/BatCal) + "v\t" + String(AnyLong) + "%");}
  if ((BatAvg > BatCritical) && (USB)) {
    // power has returned
    USB = false;
    synchLoopTimers();
  } 
  if (USB) {return;}

  BatV = getBatV(BatAvg); BatVfp = (float)BatV/10.0;

  // report WARNING state
  if (BatPc < 20) {Display_BatWarning();}

  // detect critical battery condition
  if (BatAvg <= BatCritical){
    // critical battery voltage reached
    set_LED0_New(-1);                      // turn OFF Gun LEDs
    set_LED1_New(-1);                      // turn OFF Eye LEDs
    while (true) {yield();}                 // infinite loop, everything stops in Core 1
  }
}

// ---------------------------------------------------------------------------------

void BrakeInc(float zV) {
  // Increment/decrement the current Ptch_Brake value
  if (DmB == 0) {Ptch_Brake += zV;} else {Ptch_Brake -= zV;}

  if (Ptch_Brake > 9.999) {Ptch_Brake = 9.999;}
  else if (Ptch_Brake < 0.001) {Ptch_Brake = 0.001;}
  DispDel = 1;
}

// ---------------------------------------------------------------------------------

void CalcAng() {
  // calculates the pitch, roll and turn angles of the BalanceBot BLE
  // based on Pitch: AcZ and GyY, and Roll: AcY and GyZ
  // calculate accelerometer angle
  PtchAccAng = asin((float)AcZ/8192.0)* 57.296;
  RollAccAng = asin((float)AcY/8192.0)* 57.296;
  YawAccAng  = asin((float)AcX/8192.0)* 57.296;

  // calculate gyro angles, which will be used as the true angles
  float zDiv = (float)GyTD/131068000.0;                 // nominally 0.0007629 at 10ms
  PtchGyrAng += (float)GyY * zDiv;                      // Calculate the angle travelled during this 10ms loop and add this to the angle_gyro variable
  if (PtchGyrAng > 180.0) {PtchGyrAng = 180.0;}         // limit +ve gyro swing
  else if (PtchGyrAng < -180.0) {PtchGyrAng = -180.0;}  // limit -ve gyro swing

  RollGyrAng -= (float)GyZ * zDiv;                      // Calculate the angle travelled during this 10ms loop and add this to the angle_gyro variable
  if (RollGyrAng > 180.0) {RollGyrAng = 180.0;}         // limit +ve gyro swing
  else if (RollGyrAng < -180.0) {RollGyrAng = -180.0;}  // limit -ve gyro swing

  float zYawAng = YawGyrAng;                            // note previous value
  YawGyrAng += (float)GyX * zDiv;                       // Calculate the angle travelled during this 10ms loop and add this to the angle_gyro variable
  if (YawGyrAng > 180.0) {YawGyrAng = 180.0;}           // limit +ve gyro swing
  else if (YawGyrAng < -180.0) {YawGyrAng = -180.0;}    // limit -ve gyro swing

  // eliminate drift using a small amount of AccAng
  // if not in the Ready state then force gyroangle = accangle
  if (!Ready) {PtchGyrAng = PtchAccAng;} else {PtchGyrAng = (PtchGyrAng * 0.9950) + (PtchAccAng * 0.005);}

  if (!Ready) {RollGyrAng = RollAccAng;} else {RollGyrAng = (RollGyrAng * 0.9950) + (RollAccAng * 0.005);}

  // this does not apply to YawGyrAng, as the angle cannot be corrected with gravity

  // check for corrupt angles
  if (isnan(PtchGyrAng)) {PtchGyrAng = PtchAccAng;}
  if (isnan(RollGyrAng)) {RollGyrAng = RollAccAng;}
  if (isnan(YawGyrAng)) {YawGyrAng = zYawAng;}
}

// ---------------------------------------------------------------------------------

void ChangeMainMode() {
  // called by user clicking on display monitor
}

// ---------------------------------------------------------------------------------

void CheckOrientation() {
  // Called during runPOST() or after soft reset to determine MainMode
  // take 20 readings from the MPU, in 100ms
  for (int16_t zR = 0;zR < 21; zR++) {
    readGyrosAll(); readAccelAll(); AvgAccelAll(); AvgGyrAll();
    // Serial.println("AcY=" + String(AcY));
    delay(5);
  }
  // test robots orientation
  if (AcX > +7000) {
    // the MPU6050 X-axis accelerometer is looking at the sky
    // so assume robot is resting on its front reaction wheel
    gotoDemoMode();
  } else {
    // the default mode is Test
    gotoTestMode();
  }
}

// ---------------------------------------------------------------------------------

void DbInc(float zV) {
  // Increment/decrement the current Ptch_Db value
  if (DmB == 0) {Ptch_Db += zV;} else {Ptch_Db -= zV;}

  if (Ptch_Db > 99.99) {Ptch_Db = 99.99;}
  else if (Ptch_Db < 0.000) {Ptch_Db = 0.0;}
  DispDel = 1;
}

// ---------------------------------------------------------------------------------

void DecIncRangeCountMax() {
  // change RangeCountMax value
  if (DmB == 0) {RangeCntMax++;} else {RangeCntMax--;}

  if (RangeCntMax > 99) {RangeCntMax = 99;}
  else if (RangeCntMax < 0) {RangeCntMax = 0;}
  DispDel = 1;
}

// ---------------------------------------------------------------------------------

void DecIncRangePkPk() {
  // change Range ramp rate value
  if (DmB == 0) {RangePkPk++;} else {RangePkPk--;}

  if (RangePkPk > 99) {RangePkPk = 99;}
  else if (RangePkPk < 0) {RangePkPk = 0;}
  DispDel = 1;
}

// ---------------------------------------------------------------------------------

void DecIncRangeRate() {
  // change Range ramp rate value
  if (DmB == 0) {RangeRate++;} else {RangeRate--;}

  if (RangeRate > 99) {RangeRate = 99;}
  else if (RangeRate < 0) {RangeRate = 0;}
  DispDel = 1;
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // decodes zkeyVal and excutes commands on '.'
  // every time we receive a character set response to 8ms later, just in case
  // there is another packet behind it coming from the controller
  print40ms = millis() + 6;
  if (zkeyVal == 10) {return;}
  if (zkeyVal == 13) {return;}
  keyChar = char(zkeyVal);
  switch (keyChar) {
    case '.': doCmd(); return;          // command terminator
    case '~': // connected to Monitor+ app either via USB or WiFi
      if (!DispMon) {DispMon = true; DispNext = 0; Display_Mirrored();} // display mirrored message
      DispTx = 75; return;              // apply a 3 second time-out
    case '-': cmdSgn = -1; return;      // numeric sign
    case '*': cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1; return; // ESC abort command
    case '0': extendCmdVal(0); return;
    case '1': extendCmdVal(1); return;
    case '2': extendCmdVal(2); return;
    case '3': extendCmdVal(3); return;
    case '4': extendCmdVal(4); return;
    case '5': extendCmdVal(5); return;
    case '6': extendCmdVal(6); return;
    case '7': extendCmdVal(7); return;
    case '8': extendCmdVal(8); return;
    case '9': extendCmdVal(9); return;
  }
  if (cmdMode == ' ') {
    // test for new Command Mode char?
    // only accept certain characters as valid commands, or ignore them
    // delay battery update by 10 seconds whilst receiving commands
    cmdMode = keyChar;
    switch (keyChar) {
      // check for valid mode and convert upper-case
      case 'd': cmdMode = 'D'; break;
      case 'e': cmdMode = 'E'; break;
      case 'g': cmdMode = 'G'; break;
      case 'i': cmdMode = 'I'; break;
      case 'p': cmdMode = 'P'; break;
      case 's': cmdMode = 'S'; break;
    } cmdType = ' '; cmdVal = 0;
  } else {
    // test for Command Type char?
    cmdType = keyChar;
    switch (keyChar) {
      // check for lower-case and convert to upper-case
      case 'd': cmdType = 'D'; break;
      case 'i': cmdType = 'I'; break;
      case 'm': cmdType = 'M'; break;
      case 'p': cmdType = 'P'; break;
      case 's': cmdType = 'S'; break;
      case 'v': cmdType = 'V'; break;
    }
  }
}

// --------------------------------------------------------------------------------

void displayLft() {
  // change the DispMode pointer to the previous display
  DispMode--; if (DispMode < DM_Min) {DispMode = DM_Max;}
  DispDel = 0; DispCnt = 0;
}

// --------------------------------------------------------------------------------

void displayRht() {
  // change the DispMode pointer to the next display
  DispMode++; if (DispMode > DM_Max) {DispMode = DM_Min;}
  DispDel = 0; DispCnt = 0;
}

// --------------------------------------------------------------------------------

void DM_Control_() {
  //##################################
  // Control display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if (DmX <= 50) {displayLft();}
  else {
    // has user clicked on a blue field
          if ((DmX >= 51) && (DmX <= 64) && (DmY >= 27) && (DmY <= 37)) {ChangeMainMode();}
    else if ((DmX >= 51) && (DmX <= 71) && (DmY >= 45) && (DmY <= 56)) {esp_restart();}
    else {displayRht();}
  }
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_DbBrake_() {
  //##################################
  // DeadBand & Brake display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if ((DmX >= 50) && (DmX <= 64) && (DmY >= 25) && (DmY <= 35)) {
         if ((DmX >= 50) && (DmX <= 53)) {DbInc(10.000);}
    else if ((DmX >= 54) && (DmX <= 58)) {DbInc( 1.000);}
    else if ((DmX >= 61) && (DmX <= 64)) {DbInc( 0.100);}
  } else if ((DmX >= 50) && (DmX <= 68) && (DmY >= 62) && (DmY <= 73)) {
         if ((DmX >= 50) && (DmX <= 53)) {OpBrakeInc(10.000);}
    else if ((DmX >= 54) && (DmX <= 58)) {OpBrakeInc( 1.000);}
    else if ((DmX >= 61) && (DmX <= 64)) {OpBrakeInc( 0.100);}
    else if ((DmX >= 65) && (DmX <= 68)) {OpBrakeInc( 0.010);}
  } else if ((DmX >= 50) && (DmX <= 68) && (DmY >= 80) && (DmY <= 91)) {
         if ((DmX >= 50) && (DmX <= 53)) {BrakeInc(1.000);}
    else if ((DmX >= 56) && (DmX <= 60)) {BrakeInc(0.100);}
    else if ((DmX >= 61) && (DmX <= 64)) {BrakeInc(0.010);}
    else if ((DmX >= 65) && (DmX <= 68)) {BrakeInc(0.001);}
  } else if (DmX < 50) {displayLft();}
  else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_Drive_() {
  //##################################
  // Drive display
  //##################################
  // User has clicked on a blue field in Monitor+ display
  if ((DmY >= 80) && (DmY <= 90)) {
         if ((DmX >= 26) && (DmX <= 30)) {SpTgtInc(1.000);}
    else if ((DmX >= 33) && (DmX <= 36)) {SpTgtInc(0.100);}
    else if ((DmX >= 37) && (DmX <= 41)) {SpTgtInc(0.010);}
    else if ((DmX >= 76) && (DmX <=79)) {SpdTgtInc(100.0);}
    else if ((DmX >= 80) && (DmX <=84)) {SpdTgtInc(10.0);}
    else if ((DmX >= 85) && (DmX <=88)) {SpdTgtInc(1.0);}
  } else if (DmX < 50) {displayLft();}
  else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_Filter_() {
  //##################################
  // Filter display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if ((DmY >= 53) && (DmY <= 64)) {
          if ((DmX >= 50) && (DmX <= 55)) {FilterInc(1.000);}
    else if ((DmX >= 58) && (DmX <= 61)) {FilterInc(0.100);}
    else if ((DmX >= 62) && (DmX <= 65)) {FilterInc(0.010);}
    else if ((DmX >= 66) && (DmX <=70)) {FilterInc(0.001);}
  } else if (DmX < 50) {displayLft();}
  else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_GunFire_() {
  //##################################
  // Gun Fire display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if ((DmY >= 30) && (DmY <= 40)) {
          if ((DmX >= 51) && (DmX <= 64)) {GunFire = 1;}
  } else if (DmX < 50) {displayLft();}
  else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_Motor_Stat_() {
  //##################################
  // Motor Stats display
  //##################################
  // user has clicked on a blue field in Monitor+ display
       if ((DmX >= 25) && (DmX <= 51) && (DmY >= 45) && (DmY <= 56)) {Lft_Step_Cnt = 0;}
  else if ((DmX >= 74) && (DmX <= 97) && (DmY >= 45) && (DmY <= 56)) {Rht_Step_Cnt = 0;}
  else if (DmX < 50) {displayLft();}
  else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_Motor_Test_() {
  //##################################
  // Motor Test display
  //##################################
  // Check has user clicked on a blue field?
  if ((DmX >= 66) && (DmX <= 80) && (DmY >= 7) && (DmY <= 18)) {
    Motor_Test = !Motor_Test;
    if (!Motor_Test) {
      Motors_OFF();
      if (LED0_Task == -1) {set_LED0_New(LED0_Last);}
      if (LED1_Task == -1) {set_LED1_New(LED1_Last);}
    }
  } else if ((DmX >= 27) && (DmX <= 40) && (DmY >= 26) && (DmY <= 38) && Motor_Test) {
    // toggle left enable
    Lft_En = !Lft_En; digitalWrite(StepLftEn,Lft_En);
    if (!Lft_En) {IsrRun = true; set_LED0_New(-1);  set_LED1_New(-1);}
    else {if (Lft_En && Rht_En) {IsrRun = false; set_LED0_New(LED0_Last);  set_LED1_New(LED1_Last);}}
  } else if ((DmX >= 76) && (DmX <= 90) && (DmY >= 26) && (DmY <= 38) && Motor_Test) {
    // toggle right enable
    Rht_En = !Rht_En; digitalWrite(StepRhtEn,Rht_En);
    if (!Rht_En) {IsrRun = true; set_LED0_New(-1);  set_LED1_New(-1);}
    else {if (Lft_En && Rht_En) {IsrRun = false; set_LED0_New(LED0_Last);  set_LED1_New(LED1_Last);}}
  } else if ((DmX >= 27) && (DmX <= 40) && (DmY >= 45) && (DmY <= 56) && Motor_Test) {
    // left speed
    if (DmB == 0) {
           if (Lft_Motor_Throttle >= 1000) {Lft_Motor_Throttle += 1000;}
      else if (Lft_Motor_Throttle >=  100) {Lft_Motor_Throttle += 100;}
      else {Lft_Motor_Throttle += 10;} 
    } else {
           if (Lft_Motor_Throttle >= 2000) {Lft_Motor_Throttle -= 1000;}
      else if (Lft_Motor_Throttle >=  200) {Lft_Motor_Throttle -= 100;}
      else {Lft_Motor_Throttle -= 10;} 
      if (Lft_Motor_Throttle < 10) {Lft_Motor_Throttle = 0;}
    }
  } else if ((DmX >= 76) && (DmX <= 93) && (DmY >= 45) && (DmY <= 56) && Motor_Test) {
    // right speed
    if (DmB == 0) {
           if (Rht_Motor_Throttle >= 1000) {Rht_Motor_Throttle += 1000;}
      else if (Rht_Motor_Throttle >=  100) {Rht_Motor_Throttle += 100;}
      else {Rht_Motor_Throttle += 10;} 
    } else {
           if (Rht_Motor_Throttle >= 2000) {Rht_Motor_Throttle -= 1000;}
      else if (Rht_Motor_Throttle >=  200) {Rht_Motor_Throttle -= 100;}
      else {Rht_Motor_Throttle -= 10;} 
      if (Rht_Motor_Throttle < 10) {Rht_Motor_Throttle = 0;}
    }
  } else if ((DmX >= 27) && (DmX <= 35) && (DmY >= 64) && (DmY <= 74) && Motor_Test) {
    // left Dir
    Lft_Dir = !Lft_Dir; digitalWrite(StepLftDir,Lft_Dir);
    if (Lft_Dir) {Lft_Step_Inc= -1;} else {Lft_Step_Inc= 1;} 
  } else if ((DmX >= 76) && (DmX <= 84) && (DmY >= 64) && (DmY <= 74) && Motor_Test) {
    // right Dir
    Rht_Dir = !Rht_Dir; digitalWrite(StepRhtDir,Rht_Dir);
    if (Rht_Dir) {Rht_Step_Inc= 1;} else {Rht_Step_Inc= -1;}
  } else if ((DmX >= 27) && (DmX <= 36) && (DmY >= 81) && (DmY <= 93) && Motor_Test) {
    // left M210
    if (DmB == 0) {
      switch (M210) {
        case 0b000: M210 = 0b010; break;
        case 0b001: M210 = 0b000; break;
        case 0b010: M210 = 0b001; break;
      }
    } else {
      switch (M210) {
        case 0b000: M210 = 0b001; break;
        case 0b001: M210 = 0b010; break;
        case 0b010: M210 = 0b000; break;
      }
    } Set_M210();
  } else if ((DmX >= 76) && (DmX <= 86) && (DmY >= 81) && (DmY <= 93) && Motor_Test) {
    // right M210
    if (DmB == 0) {
      switch (M210) {
        case 0b000: M210 = 0b010; break;
        case 0b001: M210 = 0b000; break;
        case 0b010: M210 = 0b001; break;
      }
    } else {
      switch (M210) {
        case 0b000: M210 = 0b001; break;
        case 0b001: M210 = 0b010; break;
        case 0b010: M210 = 0b000; break;
      }
    } Set_M210();
  }
  else if (DmX < 50) {displayLft();}
  else {displayRht();}
  DispDel = 2;  // Update after 80ms
}

// --------------------------------------------------------------------------------

void DM_MPU_Acc_() {
  //##################################
  // MPU Acc display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if (DmX < 50) {displayLft();}
  else {
    // has user clicked on a blue field
    if ((DmX >= 67) && (DmX <= 72) && (DmY >= 9) && (DmY <= 19)) {AccOffEn = !AccOffEn;}
    else {displayRht();}
  }
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_MPU_Gyr_() {
  //##################################
  // MPU Gyros display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if (DmX < 50) {displayLft();}
  else {
    // has user clicked on a blue field
    if ((DmX >= 70) && (DmX <= 74) && (DmY >= 9) && (DmY <= 19)) {GyrOffEn = !GyrOffEn;}
    else {displayRht();}
  }
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_OpMode_() {
  //##################################
  // Motor Calcs display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if ((DmX >= 50) && (DmX <= 53) && (DmY >= 25) && (DmY <= 36)) {OpModeInc();}
  else if (DmX < 50) {displayLft();} else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_PtchPID_() {
  //##################################
  // Pitch PID display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if (DmX < 50) {displayLft();}
  else if ((DmX < 69) && (DmY >= 25) && (DmY <= 36)) {
    // user has clicked on P-gain field
          if ((DmX >= 50) && (DmX <= 53)) {PtchPinc(10.0);}
    else if ((DmX >= 54) && (DmX <= 57)) {PtchPinc(1.0);}
    else if ((DmX >= 61) && (DmX <= 64)) {PtchPinc(0.1);}
    else if ((DmX >= 65) && (DmX <= 68)) {PtchPinc(0.01);}
  } else if ((DmX < 69) && (DmY >= 44) && (DmY <= 54)) {
    // user has clicked on I-gain field
          if ((DmX >= 50) && (DmX <= 53)) {PtchIinc(10.0);}
    else if ((DmX >= 54) && (DmX <= 57)) {PtchIinc(1.0);}
    else if ((DmX >= 61) && (DmX <= 64)) {PtchIinc(0.1);}
    else if ((DmX >= 65) && (DmX <= 68)) {PtchIinc(0.01);}
  } else if ((DmX >= 78) && (DmX <= 96) && (DmY >= 44) && (DmY <= 54)) {
    // user has clicked on I-max field
          if ((DmX >= 78) && (DmX <= 81)) {PtchIMinc(10.0);}
    else if ((DmX >= 82) && (DmX <= 85)) {PtchIMinc(1.0);}
    else if ((DmX >= 88) && (DmX <= 92)) {PtchIMinc(0.1);}
    else if ((DmX >= 93) && (DmX <= 96)) {PtchIMinc(0.01);}
  } else if ((DmX < 69) && (DmY >= 62) && (DmY <= 73)) {
    // user has clicked on D-gain field
          if ((DmX >= 50) && (DmX <= 53)) {PtchDinc(10.0);}
    else if ((DmX >= 54) && (DmX <= 57)) {PtchDinc(1.0);}
    else if ((DmX >= 61) && (DmX <= 64)) {PtchDinc(0.1);}
    else if ((DmX >= 65) && (DmX <= 68)) {PtchDinc(0.01);}
  } else if ((DmX < 69) && (DmY >= 79) && (DmY <= 91)) {
    // user has clicked on StartAng field
         if ((DmX >= 54) && (DmX <= 57)) {PtchSinc(1.0);}
    else if ((DmX >= 61) && (DmX <= 64)) {PtchSinc(0.1);}
    else if ((DmX >= 65) && (DmX <= 68)) {PtchSinc(0.01);}
  } else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_Range_() {
  //##################################
  // Range display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if ((DmX >= 32) && (DmX <= 67) && (DmY >= 9) && (DmY <= 31)) {
    if (RangeEn) {
      VL53L1X_OFF(); set_LED0_New(2);
    } else {
      VL53L1X_ON(); set_LED0_New(5);
    }
  }
  else if (RangeEn) {
    // is the user clicking on blue fields
         if ((DmX >= 24) && (DmX <= 31) && (DmY >= 65) && (DmY <= 76)) {DecIncRangeCountMax();}
    else if ((DmX >= 24) && (DmX <= 31) && (DmY >= 80) && (DmY <= 91)) {DecIncRangePkPk();}
    else if ((DmX >= 88) && (DmX <= 95) && (DmY >= 80) && (DmY <= 91)) {DecIncRangeRate();}
    else if (DmX < 50) {displayLft();} else {displayRht();}
  }
  else if (DmX < 50) {displayLft();} else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void DM_SafeMode_() {
  //##################################
  // SafeMode display
  //##################################
  // user has clicked on a blue field in Monitor+ display
  if ((DmX >= 74) && (DmX <= 97) && (DmY >= 44) && (DmY <= 54)) {
    // user has clicked on SbRollSp value field
         if ((DmX >= 74) && (DmX <= 78)) {PtchSbInc(1.000);}
    else if ((DmX >= 81) && (DmX <= 84)) {PtchSbInc(0.100);}
    else if ((DmX >= 85) && (DmX <= 88)) {PtchSbInc(0.010);}
    else if ((DmX >= 89) && (DmX <= 93)) {PtchSbInc(0.001);}
    else if ((DmX >= 94) && (DmX <= 97)) {PtchSbInc(0.0001);}
  } else if (DmX < 50) {displayLft();} else {displayRht();}
  DispDel = 1;  // Update after 40ms
}

// --------------------------------------------------------------------------------

void doCmd() {
  // A '.' character has been received, so execute command if recognised.
  // Supported commands (mainly from Monitor+) are:
  // .        - report P,I,D gains
  // Dn.      - D-gain enable, 1 = ON, 0 = OFF
  // DMxxyy.  - display monitor left mouse click, xx = 0-99, yy = 0-99
  // DM-.     - display monitor mouse up
  // DVnn.    - D-gain value * 100
  // G.       - report Scope data for graphing app
  // GDnn.    - decrement the ScpPnt 0-3 to table case value by nn, for custom GP0 mode
  // GE.      - end of Scope graphing mode
  // GInn.    - increment the ScpPnt 0-3 to table case value by nn, for custom GP0 mode
  // GPnn.    - nn points to data to be sent
  // GSnn.    - set a pointer 0-3 to ScpPnt0-3
  // In.      - I-gain enable, 1 = ON, 0 = OFF
  // IVnn.    - I-gain value * 100
  // PVnn.    - P-gain value * 1000
  //
  switch (cmdMode) {
    case 'D':
      switch (cmdType) {
        case 'M':
          if (cmdSgn < 0) {}  // mouse click released
          else {
            if (cmdVal >= 10000) {
              // this is a Display Monitor mouse right-click
              // determine whether it is a left or right button click
              cmdVal /= 10000; cmdVal --; DmB = 1;
              DmY = cmdVal % 100; DmX = cmdVal/100;
            } else {
              // this is a Display Monitor mouse left-click
              DmY = cmdVal % 100; DmX = cmdVal/100; DmB = 0;
            }
                 if (DispMode == DM_Control) {DM_Control_();}
            else if (DispMode == DM_DbBrake) {DM_DbBrake_();}
            else if (DispMode == DM_Drive) {DM_Drive_();}
            else if (DispMode == DM_Filter) {DM_Filter_();}
            else if (DispMode == DM_GunFire) {DM_GunFire_();}
            else if (DispMode == DM_Motor_Stat) {DM_Motor_Stat_();}
            else if (DispMode == DM_Motor_Test) {DM_Motor_Test_();}
            else if (DispMode == DM_MPU_Acc) {DM_MPU_Acc_();}
            else if (DispMode == DM_MPU_Gyros) {DM_MPU_Gyr_();}
            else if (DispMode == DM_OpMode) {DM_OpMode_();}
            else if (DispMode == DM_PtchPID) {DM_PtchPID_();}
            else if (DispMode == DM_Range) {DM_Range_();}
            else if (DispMode == DM_SafeMode) {DM_SafeMode_();}
            else {
              // this is not a special 'clickable' display, so simply change it
              if (DmX < 50) {displayLft();  // left side click
              } else {displayRht(); }       // right side click
            }
          } break;
      } break;
    case 'G':
      // Scope graphing app report commands
      switch (cmdType) {
        case ' ': 
          // report the same set of values twice as WiFi data can easily be
          // corrupted. Therefore two matching consequitives messages must be
          // received for the graphing function to work
          // if in Display Monitor mode then switch that off
          if (DispMon) {DispMon = false; PrintTx = ""; DispTx = -1;}
          ScopeReport(GPnn); ScopeReport(GPnn);
          break;
        case 'D': // decrement the ScpPnt 0-3 to table case value, for custom GP0 mode
               if (ScpPP == 0) {ScpPnt0 -= cmdVal; if (ScpPnt0 < 0) {ScpPnt0 += (ScpDpth + 1);}}
          else if (ScpPP == 1) {ScpPnt1 -= cmdVal; if (ScpPnt1 < 0) {ScpPnt1 += (ScpDpth + 1);}}
          else if (ScpPP == 2) {ScpPnt2 -= cmdVal; if (ScpPnt2 < 0) {ScpPnt2 += (ScpDpth + 1);}}
          else if (ScpPP == 3) {ScpPnt3 -= cmdVal; if (ScpPnt3 < 0) {ScpPnt3 += (ScpDpth + 1);}}
          break;
        case 'E': // GE. enter monitor mode when dropping out of graphing mode
          // return to Display Monitor mode if it was previously active
          if (DispTx < 0) {DispMon = true; PrintTx = ""; DispTx = 75; DispDel = 0; DispCnt = 0;}
          break;
        case 'I': // increment the ScpPnt 0-3 to table case value, for custom GP0 mode
               if (ScpPP == 0) {ScpPnt0 += cmdVal; if (ScpPnt0 > ScpDpth) {ScpPnt0 -= (ScpDpth + 1);}}
          else if (ScpPP == 1) {ScpPnt1 += cmdVal; if (ScpPnt1 > ScpDpth) {ScpPnt1 -= (ScpDpth + 1);}}
          else if (ScpPP == 2) {ScpPnt2 += cmdVal; if (ScpPnt2 > ScpDpth) {ScpPnt2 -= (ScpDpth + 1);}}
          else if (ScpPP == 3) {ScpPnt3 += cmdVal; if (ScpPnt3 > ScpDpth) {ScpPnt3 -= (ScpDpth + 1);}}
          break;
        case 'P': // GPnn. sets the reporting mode and triggers title reporting
          if (DispMon) {DispMon = false; PrintTx = ""; DispTx = -1;}
          GPnn = cmdVal; GP_Title = true;
          PrintTx += "GP" + String(GPnn) + "\n";
          break;
        case 'S': // set a pointer 0-3 to ScpPnt0-3
          ScpPP = cmdVal; break;
      } break;
  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
}

// --------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = (cmdVal * 10) + zVal;  
}

// --------------------------------------------------------------------------------

void EyeLED_Tasks() {
  // Called from the Core0 loop0() every 20ms (50Hz) to perform a range of LED tasks
  // LED_Task functions are:
  // -1 - clear EyeLED array to OFF and set flags
  //  0 - clear flag and do nothing
  //  1 - random moving eyes, left and right
  // 10 - bright random colours during Wii control connect/disconnect
  if (!sw0State || !sw1State) {
    // one or both switches have been pressed, so indicate this
    fill_solid(Eyes_LED, 7, CRGB::Black);
    if (!sw0State) {Eyes_LED[1].setRGB(0,0,128);}
    if (!sw1State) {Eyes_LED[5].setRGB(0,0,128);}
     LED1_Run = false; LEDshow1 = true; LED1_Del = 3; return;
  }
  if (LED1_Flash) {
    // bright white flash
    LED1_Del = LED1_Flash; LED1_Flash = 0;
    Eyes_LED[0].setRGB(128,128,128); Eyes_LED[1].setRGB(128,128,128); Eyes_LED[2].setRGB(128,128,128);
    Eyes_LED[3].setRGB(128,128,128); Eyes_LED[4].setRGB(128,128,128);
     LED1_Run = false; LEDshow1 = true;
  }

  if (GunFire) {return;}

  if (LED1_Del > 0) {LED1_Del--; LED1_Run = false; return;}

  // perform specified task
  switch(LED1_Task) {
    case -1: fill_solid(Eyes_LED, 7, CRGB::Black); LED1_Run = false; LED1_Run = false; LEDshow1 = true; LED1_Del = 500; break;
    case  0: LED1_Run = false; break;
    case  1: EyeLED_Tasks1(); break;  // moving eyes
    case 10: // random LED colours for Wii connect/disconnect
      for (int16_t zP = 0;zP < NumLEDs;zP++) {
        GetRndCol(32); Eyes_LED[zP].setRGB(Col_R,Col_G,Col_B);
      }
      LED1_Run = false; LEDshow1 = true; LED1_Del = 6; break;
  }
}

// --------------------------------------------------------------------------------

void EyeLED_Tasks1() {
  // Random moving eyes, which blink on occasions
  // Eye colour is dependant on SafeModes and robot conditions.
  if (SafeMode <= 0) {
    if (WiFiMode) {Temp_LED[0].setRGB(  0,128,  0);}  // green eyes for WiFi mode
             else {Temp_LED[0].setRGB(  0,  0,128);}  // blue eyes for BLE mode
  } else {Temp_LED[0].setRGB(128,  0,  0);} // red eyes
  switch(LED1_SubTsk) {
    case 0: //look forward
      fill_solid(Eyes_LED, 7, CRGB::Black);
      Eyes_LED[2] = Temp_LED[0]; Eyes_LED[4] = Temp_LED[0]; Bright1 = 0;
      LED1_Pnt = 2; LED1_SubTsk++; break;
    case 1: //increase brightness, then pause
      LED1_Del = 1;
      if (Bright1 < 255) {Bright1 += 5;} else {LED1_SubTsk++; LED1_Del = 20 + random(60);}
      break;
    case 2: // choose to look elsewhere
      LED1_Cnt = random(5);
      if (LED1_Cnt != LED1_Pnt) {
        if (LED1_Pnt < LED1_Cnt) {LED1_Pnt++;} else {LED1_Pnt--;} // make the 1st move
        fill_solid(Eyes_LED, 7, CRGB::Black);
        Eyes_LED[LED1_Pnt] = Temp_LED[0]; Eyes_LED[LED1_Pnt + 2] = Temp_LED[0];
        LED1_SubTsk++; LED1_Del = 6;
      } break;
    case 3: // move eyes
      LED1_Del = 6; if (LED1_Pnt == LED1_Cnt) {LED1_SubTsk++;}        // reached position
      else if (LED1_Pnt < LED1_Cnt) {LED1_Pnt++;} else {LED1_Pnt--;}  // move towards current target
      fill_solid(Eyes_LED, 7, CRGB::Black);
      Eyes_LED[LED1_Pnt] = Temp_LED[0]; Eyes_LED[LED1_Pnt + 2] = Temp_LED[0];
      break;
    case 4: // delay and blink
      if (random(10) < 2) {LED1_SubTsk++; LED1_Del = 2 + random(5);} else {LED1_SubTsk = 6;}
      break;
    case 5: // blink
      Bright1 = 0; LED1_Del = 4; LED1_SubTsk++;
      break;
    case 6: // set a pause
      Bright1 = 255; LED1_Del = 20 + random(60); LED1_SubTsk = 2; break;
  } LED1_Run = false; LEDshow1 = true;
}

// --------------------------------------------------------------------------------

void FilterInc(float zInc) {
  // inc/dec LpFRc filter coefficient
  if (DmB != 0) {
    // mouse right-click
    LpFc -= zInc;
  } else {
    // mouse left-click
    LpFc += zInc;
  }
  if (LpFc > 1.000) {LpFc = 1.000;}
  else if (LpFc < 0.0010) {LpFc = 0.0010;}
}

// --------------------------------------------------------------------------------

int16_t getBatV(int32_t zAvg) {
  // Determine the battery voltage, and percentage, by mapping the ADC readings onto threshold Voltages
  // These threshold voltages were determined using a multimeter, to calibrate the ESP32's ADC
  // if the voltage exceeds BatMax, the estimated voltage is displayed, but battery capacity will be limited to 100%
       if (zAvg > BatMax)  {BatV = map(zAvg,3402,3841,114,123); BatPc = 100;} // >= 100%
  else if (zAvg > BatHigh) {BatV = map(zAvg,3402,3841,114,123); BatPc = map(zAvg,3402,3841,80,100);}  // >= 80%
  else if (zAvg > BatWarn) {BatV = map(zAvg,3175,3402,105,114); BatPc = map(zAvg,3175,3402,20, 80);}  // >= 20%
  else {BatV = map(zAvg,2895,3175,100,105); BatPc = map(zAvg,2895,3175,0,20);}  // < 20%
  return BatV;
}

// --------------------------------------------------------------------------------

String getBinary(byte zB) {
  // returns an 8-character string for zB
  if (zB & 0b10000000) {Any$  = "1";} else {Any$  = "0";}
  if (zB & 0b01000000) {Any$ += "1";} else {Any$ += "0";}
  if (zB & 0b00100000) {Any$ += "1";} else {Any$ += "0";}
  if (zB & 0b00010000) {Any$ += "1";} else {Any$ += "0";}
  if (zB & 0b00001000) {Any$ += "1";} else {Any$ += "0";}
  if (zB & 0b00000100) {Any$ += "1";} else {Any$ += "0";}
  if (zB & 0b00000010) {Any$ += "1";} else {Any$ += "0";}
  if (zB & 0b00000001) {Any$ += "1";} else {Any$ += "0";}
  return Any$;
}

// --------------------------------------------------------------------------------

String getHmsTime(unsigned long zMillis) {
  // returns a 'time' string in Hrs:Min:sec base on zMillis
  String zTime$ = "";
  long zU = zMillis/3600000; zMillis = zMillis%3600000;
  if (zU > 0) {zTime$ = String(zU) + ":";}
  zU = zMillis/60000; zMillis = zMillis%60000;
  if (zU < 10) {zTime$ += "0";}
  zTime$ += String(zU);
  zU = zMillis/1000;
  if (zU < 10) {zTime$ += ":0";} else {zTime$ += ":";}
  zTime$ += String(zU);
  return zTime$;
}

// --------------------------------------------------------------------------------

String GetM210(uint8_t zM210) {
  // return a string for display porposes
  if ((zM210 & 0b100) > 0) {Any$ = "1";} else {Any$ = "0";}
  if ((zM210 & 0b010) > 0) {Any$ += "1";} else {Any$ += "0";}
  if ((zM210 & 0b001) > 0) {Any$ += "1";} else {Any$ += "0";}
  return Any$;
}

// --------------------------------------------------------------------------------

String getNum00(int16_t zV) {
  // returns a two character number 00 - 99
  if (zV < 10) {return "0" + String(zV);}
  else {return String(zV);}
}

// --------------------------------------------------------------------------------

void GetRndCol(byte zB) {
  // return a random colour of strength zB
  if (zB < 2) {zB = 2;}
  switch(random(6)) {
    case 0: Col_R =   zB; Col_G =    0; Col_B =    0; break;  // red
    case 1: Col_R = zB/2; Col_G = zB/2; Col_B =    0; break;  // yellow
    case 2: Col_R =    0; Col_G =   zB; Col_B =    0; break;  // green
    case 3: Col_R =    0; Col_G = zB/2; Col_B = zB/2; break;  // cyan
    case 4: Col_R =    0; Col_G =    0; Col_B =   zB; break;  // blue
    case 5: Col_R = zB/2; Col_G =    0; Col_B = zB/2; break;  // magenta
  }
}

// --------------------------------------------------------------------------------

void gotoDemoMode() {
  // Initialises conditions for Demo Mode, if needed
  // Serial.println(F("Demo Mode"));
  MainMode = 0;   // set main loop flag for DEMO mode
  resetDMs();     // reset the displaymode references
  // setDefaultsB(); // reset some variables
  // // if (I2C_MPU) {MPU_6050_Initialise();} // ensure recent crash hasn't affected the MPU micro
  if (!USB) {
  //   // on 12v battery power
  //   Gear = 0;                   // speed control flag, default = 0
  //   PIDtask = 0;                // start with PID calculation sequence OFF
  //   Ptchen = true;              // enable Pitch PID calculations & safe modes
  //   PtchGo = false;             // ensure Pitch go flag is reset
  //   PtchSafeMode = -1;          // initialise Pitch safe mode
  //   Rollen = true;              // enable Roll PID calculations & safe modes
  //   RollGo = false;             // ensure Roll go flag is reset
  //   RollSafeMode = -1;          // initialise Roll safe mode
  //   SbPtchSpInc = SbPtchSpIncI; // set initial auto-adjust increment
  //   SbRollSpInc = SbRollSpIncI; // set initial auto-adjust increment
  //   TurnTask = 0;               // resaet Turn state machine
  //   set_LED_Task(1);            // Demo Safe Mode LEDs
  //   AutoOFF = 100;              // set auto remove motor power after 2 seconds
  //   changeGear(Gear);           // set speeds
  //   Display_Text_2x16("DEMO","Mode",50);
  } else {
  //   // on USB power only
  //   PIDtask = 0;                // turn off PID calculation sequence
  //   Ptchen = false;             // disable Pitch safe mode & PID calculations
  //   Rollen = false;             // disable Roll safe mode & PID calculations
  //   set_LED_Task(4);            // green blink
  //   Display_Text_2x16("DEMO","USB Pwr",50);
  }
  Task8ms = 0;
  synchLoopTimers();
}

// --------------------------------------------------------------------------------

void gotoTestMode() {
  // Initialises conditions for Test Mode, as needed
  // Serial.println(F("Test Mode"));
  MainMode = 1;   // set main loop flag for TEST mode
  resetDMs();     // reset the displaymode references
  // PIDtask = 0;                // turn off PID calculation sequence
  // Ptchen = false;             // disable Pitch safe mode
  // PtchGo = false;             // ensure Pitch go flag is reset
  // Rollen = false;             // disable Roll safe mode
  // RollGo = false;             // ensure Roll go flag is reset
  // MainMode = 1;               // set main loop flag for Demo mode
  // TestMode = -1;                 // start in default state
  Task8ms = 0;
  synchLoopTimers();
}

// --------------------------------------------------------------------------------

void GunLED_Fire() {
  // Runs through a firing sequence for both guns, then cancels itself.
  // Also controls sounds during firing.
  if (ToneFreq) {TonePlay(ToneFreq - 50);}
  if (LED0_Del > 0) {LED0_Del--; LED0_Run = false; return;} // delay by skipping tasks

  // Set chest colour red
  switch(GunFire) {
    case  1:
      Guns_LED[0].setRGB(255,255,255); TonePlay(2093); LED0_Del = 5;
      Guns_LED[2].setRGB(8,0,0); Guns_LED[3] = Guns_LED[2]; Guns_LED[4] = Guns_LED[2];
      break;
    case  2: Guns_LED[0].setRGB(  0,  0,  0); ToneDetach(); LED0_Del = 5; break;
    case  3: Guns_LED[0].setRGB(255,255,255); TonePlay(2093); LED0_Del = 5; break;
    case  4: Guns_LED[0].setRGB(  0,  0,  0); ToneDetach(); LED0_Del = 5; break;
    case  5: Guns_LED[0].setRGB(255,255,255); TonePlay(2093); LED0_Del = 5; break;
    case  6: Guns_LED[0].setRGB(  0,  0,  0); LED0_Del = 5; break;
    case  7: Guns_LED[0].setRGB(128,128,  0); LED0_Del = 5; break;
    case  8: Guns_LED[0].setRGB( 64, 16,  0); ToneDetach(); LED0_Del = 5; break;
    case  9: Guns_LED[0].setRGB( 32,  0,  0); LED0_Del = 5; break;
    case 10: Guns_LED[0].setRGB(  8,  0,  8); LED0_Del = 5; break;
    case 11: Guns_LED[0].setRGB(  2,  0,  6); LED0_Del = 5; break;
    case 12: Guns_LED[0].setRGB(  0,  0,  4); LED0_Del = 10; break;
    case 13: Guns_LED[0].setRGB(  0,  0,  2); LED0_Del = 10; break;
    case 14: Guns_LED[0].setRGB(  0,  0,  1); LED0_Del = 10; break;
    case 15: Guns_LED[0].setRGB(  0,  0,  0); LED0_Del = 5; break;
    default: // reached the end of the sequence
      GunFire = 0; LED0_Del = 0;  LED1_Del = 0; return;
  }
  Guns_LED[1] = Guns_LED[0]; Bright0 = 255;
  fill_solid(Eyes_LED, 7, CRGB::Black);
  if (GunFire < 7) {Eyes_LED[2].setRGB(255,0,0);} else {Eyes_LED[2].setRGB(255/(2*GunFire),0,0);} 
  Eyes_LED[4]= Eyes_LED[2];
  LED0_Run = false; LED1_Run = false;
  LEDshow0 = true; LEDshow1 = true; GunFire++;
}

// --------------------------------------------------------------------------------

void GunLED_Tasks() {
  // Called from the Core0 loop0() every 20ms (50Hz) to perform a range of LED tasks.
  // LED_Task functions are:
  // -1 - turn OFF all GunLEDs, then set flags to display
  //  0 - clear flags, do nothing
  //  1 - intro sequence, controls both gun and eye LEDs
  //  2 - battery monitoring (default)
  //  3 - SafeMode count monitoring, heart beat and angle
  //  4 - show current Gear, as a binary number
  //  5 - display Range as an LED strip, also affects eye LEDs
  // 10 - bright random colours during Wii control connect/disconnect

  if (!sw0State || !sw1State) {
    // one or both switches have been pressed, so indicate this
    fill_solid(Guns_LED, 7, CRGB::Black);
    LED0_Run = false; LEDshow0 = true; LED0_Del = 3; return;
  }
  if (LED0_Flash) {
    // bright white flash
    LED0_Del = LED0_Flash; LED0_Flash = 0;
    Guns_LED[0].setRGB(128,128,128); Guns_LED[1].setRGB(128,128,128); Guns_LED[2].setRGB(128,128,128);
    Guns_LED[3].setRGB(128,128,128); Guns_LED[4].setRGB(128,128,128);
    LED0_Run = false; LEDshow0 = true;
  }

  if (GunFire) {GunLED_Fire(); return;}

  if (LED0_Del > 0) {LED0_Del--; LED0_Run = false; return;}

  // perform specified task
  switch(LED0_Task) {
    case -1: fill_solid(Guns_LED, 7, CRGB::Black); LED0_Run = false; LEDshow0 = true; LED0_Del = 500; break;
    case  0: LED0_Run = false; break;
    case  1: GunLED_Task1(); break;   // intro sequence, controls both gun and eye LEDs
    case  2: GunLED_Task2(); break;   // battery monitoring (default)
    case  3: GunLED_Task3(); break;   // SafeMode count monitoring, heart beat and angle
    case  4: GunLED_Task4(); break;   // show current Gear, as a binary number
    case  5: GunLED_Task5(); break;   // display Range as an LED strip, also affects eye LEDs
    case 10: // random LED colours for Wii connect/disconnect
      for (int16_t zP = 0;zP < NumLEDs;zP++) {
        GetRndCol(32); Guns_LED[zP].setRGB(Col_R,Col_G,Col_B);
      } LED0_Run = false; LEDshow0 = true; LED0_Del = 6;
      break;
  }
}

// --------------------------------------------------------------------------------

void GunLED_Task1() {
  // Intro sequence, controls all LEDs and sets the Ready flag once done
  switch (LED0_SubTsk) {
    case 0: // initialise
      set_LED1_Task(0);   // disable LED1 tasks
      LED0_SubTsk++; break;
    case 1: // rainbow colour sequence
      // rotate LEDs up and outwards
      Eyes_LED[0] = Eyes_LED[1]; Eyes_LED[6] = Eyes_LED[5];
      Eyes_LED[1] = Eyes_LED[2]; Eyes_LED[5] = Eyes_LED[4];
      Eyes_LED[2] = Eyes_LED[3]; Eyes_LED[4] = Eyes_LED[3];
      Eyes_LED[3] = Guns_LED[2]; Guns_LED[2] = Guns_LED[3];
      Guns_LED[3] = Guns_LED[4]; Guns_LED[4] = Guns_LED[1];
      // select the next colour
      switch (LED0_Pnt) {
        case  0: Guns_LED[0].setRGB(128,  0,  0); break; // red
        case  1: Guns_LED[0].setRGB( 96, 32,  0); break; // yellow
        case  2: Guns_LED[0].setRGB( 64, 64,  0); break; // yellow
        case  3: Guns_LED[0].setRGB( 32, 96,  0); break; // green
        case  4: Guns_LED[0].setRGB(  0,128,  0); break; // green
        case  5: Guns_LED[0].setRGB(  0, 96, 32); break; // cian
        case  6: Guns_LED[0].setRGB(  0, 64, 64); break; // cian
        case  7: Guns_LED[0].setRGB(  0, 32, 96); break; // blue
        case  8: Guns_LED[0].setRGB(  0,  0,128); break; // blue
        case  9: Guns_LED[0].setRGB( 32,  0, 96); break; // magenta
        case 10: Guns_LED[0].setRGB( 64,  0, 64); break; // magenta
        case 11: Guns_LED[0].setRGB( 96,  0, 32); break; // magenta
      } Guns_LED[1] = Guns_LED[0];  // set the guns the same
      LED0_Pnt++; if (LED0_Pnt > 11) {LED0_Pnt = 0; LED0_Cnt++;}
      LED0_Run = false; LEDshow0 = true; LED1_Run = false; LEDshow1 = true; LED0_Del = 3;
      if (LED0_Cnt > 5) {Bright0 = 240; LED0_SubTsk++;}
      break;
    case 2: // fade out the LEDs
      if (Bright0 > 0) {
        Bright0 -= 20; Bright1 = Bright0; LED0_Del = 1;
        LED0_Run = false; LEDshow0 = true; LED1_Run = false; LEDshow1 = true;
      } else {
        Guns_LED[0].setRGB(0,0,0); Guns_LED[1] = Guns_LED[0];
        set_LED0_New(2); set_LED1_New(1); Ready = true; Beep = -100;
      } break;
  }
}

// --------------------------------------------------------------------------------

void GunLED_Task2() {
  // battery monitoring, pulsing bar graph
  switch (LED0_SubTsk) {
    case 0: // initialise
      Bright0 = 0; LED0_Inc = 25; LED0_SubTsk++; break;
    case 1:
      Guns_LED[2].setRGB(0,0,0); Guns_LED[3]= Guns_LED[2]; Guns_LED[4] = Guns_LED[2];
          if (USB) {Guns_LED[4].setRGB( 64,  0, 64);}  // USB is purple
      else if (BatPc >= 90) {Guns_LED[2].setRGB(  0,128,  0); Guns_LED[3] = Guns_LED[2]; Guns_LED[4] = Guns_LED[2];}
      else if (BatPc >= 75) {Guns_LED[4].setRGB(  0,128,  0); Guns_LED[3] = Guns_LED[4]; Guns_LED[2].setRGB( 32, 32,  0);}
      else if (BatPc >= 65) {Guns_LED[4].setRGB(128,128,  0); Guns_LED[3] = Guns_LED[4];}
      else if (BatPc >= 50) {Guns_LED[4].setRGB(128,128,  0); Guns_LED[3].setRGB(64, 32,  0);}
      else if (BatPc >= 25) {Guns_LED[4].setRGB( 64, 32,  0);}
      else {Guns_LED[4].setRGB( 64,  0,  0);}
      Bright0 += LED0_Inc;
      if (Bright0 >= 250) {LED0_Inc = -50;} else if (Bright0 == 0) {LED0_Inc = 25;}
      LED0_Run = false; LEDshow0 = true; LED0_Del = 10;
      break;
  }
}

// --------------------------------------------------------------------------------

void GunLED_Task3() {
  // SafeMode count monitoring, heart beat and angle
  switch (SafeMode) {
    case 0: // SafeCnt counting
      // nose LEDs are blue, counting down
           if (SafeCnt < (SafeCntMax1)) {
             if (!USB) {
              Guns_LED[3].setRGB(  0,  0,128); Guns_LED[4].setRGB(  0,  0,128);
              Col_B = map(SafeCnt,0,SafeCntMax1,128,0); Guns_LED[2].setRGB(  0,  0,Col_B);
             } else {
              Guns_LED[3].setRGB( 64,  0, 64); Guns_LED[4].setRGB( 64,  0, 64);
              Col_B = map(SafeCnt,0,SafeCntMax1,64,0); Guns_LED[2].setRGB(Col_B,  0,Col_B);
             }
           }
      else if (SafeCnt < ((SafeCntMax * 2)/3)) {
            Guns_LED[2].setRGB(  0,  0,  0);
            if (!USB) {
              Guns_LED[4].setRGB(  0,  0,128);
              Col_B = map(SafeCnt,SafeCntMax1,SafeCntMax2,128,0); Guns_LED[3].setRGB(  0,  0,Col_B);
            } else {
              Guns_LED[4].setRGB( 64,  0,  64);
              Col_B = map(SafeCnt,SafeCntMax1,SafeCntMax2,64,0); Guns_LED[3].setRGB(Col_B,  0,Col_B);
            }
       }
      else {
            Guns_LED[2].setRGB(  0,  0,  0); Guns_LED[3].setRGB(  0,  0,  0);
            if (!USB) {
              Col_B = map(SafeCnt,SafeCntMax2,SafeCntMax,128,0); Guns_LED[4].setRGB(  0,  0,Col_B);
            } else {
              Col_B = map(SafeCnt,SafeCntMax2,SafeCntMax,64,0); Guns_LED[4].setRGB(Col_B,  0,Col_B);
            }
      } 
      LED0_Run = false; LEDshow0 = true; break;
    case 1: // heart pulse and angle
      Guns_LED[2].setRGB(  0,  0,  0); Guns_LED[3].setRGB(  0,  0,  0);
      AnyInt = (int)PtchGyrAng;
      Col_R = map(AnyInt,0,90,24,196); AnyInt = map(AnyInt,0,90,25,50);
      LED0_Cnt++; if (LED0_Cnt > AnyInt) {LED0_Cnt = 0;}
           if (LED0_Cnt ==  0) {Guns_LED[4].setRGB(Col_R,  0,  0); LEDshow0 = true;}
      else if (LED0_Cnt ==  5) {Guns_LED[4].setRGB(  0,  0,  0); LEDshow0 = true;}
      else if (LED0_Cnt == 10) {Guns_LED[4].setRGB(Col_R,  0,  0); LEDshow0 = true;}
      else if (LED0_Cnt == 15) {Guns_LED[4].setRGB(  0,  0,  0); LEDshow0 = true;}
      LED0_Run = false; break;
    case 3: // return to battery monitoring
      set_LED0_New(2); break;
  }
}

// --------------------------------------------------------------------------------

void GunLED_Task4() {
  // display GEAR value as binary digits in teh mouth LEDs
  Guns_LED[2].setRGB(0,0,0); Guns_LED[3]= Guns_LED[2]; Guns_LED[4] = Guns_LED[2];
  switch (Gear) {
    case 1:
      Guns_LED[4].setRGB(128,64,  0); break;
    case 2:
      Guns_LED[3].setRGB(128,64,  0); break;
    case 3:
      Guns_LED[3].setRGB(128,64,  0); Guns_LED[4] = Guns_LED[3]; break;
    case 4:
      Guns_LED[2].setRGB(128,64,  0); break;
  }
  // show, then return to previous LED task
  Bright0 = 255; LED0_Del = 20;
  LED0_Run = false; LEDshow0 = true; LED0_Task = LED0_Last;
}

// --------------------------------------------------------------------------------

void GunLED_Task5() {
  // Display Range value as a strips of LEDs
  // Start by clearing both strips
  fill_solid(Eyes_LED, 7, CRGB::Black); fill_solid(Guns_LED, 7, CRGB::Black);
  LED0_Pnt = map(Range,LtofMin,150,6,0);
  if (LED0_Pnt > 6) {LED0_Pnt = 6;}
  switch (LED0_Pnt) {
    case 0:
      Guns_LED[4].setRGB(128,64,  0); break;
    case 1:
      Guns_LED[4].setRGB(144,72,  0); Guns_LED[3] = Guns_LED[4]; break;
    case 2:
      Guns_LED[4].setRGB(160,80,  0); Guns_LED[3] = Guns_LED[4];
      Guns_LED[2] = Guns_LED[4]; break;
    case 3:
      Guns_LED[4].setRGB(176,88,  0); Guns_LED[3] = Guns_LED[4];
      Guns_LED[2] = Guns_LED[4]; Eyes_LED[3] = Guns_LED[4]; break;
    case 4:
      Guns_LED[4].setRGB(192,96,  0); Guns_LED[3] = Guns_LED[4];
      Guns_LED[2] = Guns_LED[4]; Eyes_LED[3] = Guns_LED[4];
      Eyes_LED[2] = Guns_LED[4]; Eyes_LED[4] = Guns_LED[4]; break;
    case 5:
      Guns_LED[4].setRGB(208,104,  0); Guns_LED[3] = Guns_LED[4];
      Guns_LED[2] = Guns_LED[4]; Eyes_LED[3] = Guns_LED[4];
      Eyes_LED[2] = Guns_LED[4]; Eyes_LED[4] = Guns_LED[4];
      Eyes_LED[1] = Guns_LED[4]; Eyes_LED[5] = Guns_LED[4]; break;
    case 6:
      Guns_LED[4].setRGB(224,112,  0); Guns_LED[3] = Guns_LED[4];
      Guns_LED[2] = Guns_LED[4]; Eyes_LED[3] = Guns_LED[4];
      Eyes_LED[2] = Guns_LED[4]; Eyes_LED[4] = Guns_LED[4];
      Eyes_LED[1] = Guns_LED[4]; Eyes_LED[5] = Guns_LED[4];
      Eyes_LED[0] = Guns_LED[4]; Eyes_LED[6] = Guns_LED[4]; break;
  }
  // Show both strips, then return to previous LED task
  Bright0 = 255; LED0_Del = 2;
  LED0_Run = false; LEDshow0 = true; LED1_Run = false; LEDshow1 = true;
}

// --------------------------------------------------------------------------------

void initBatteryRead() {
  // called during runPOST() or on a soft reset
  // take 100 battery readings over a 100ms period to get an average
  for (int zI = 0;zI < 100;zI++) {BatVol+= analogRead(BatPin); delay(1);}
  BatVol/= 100;

  BatAvg = BatVol; BatSum = BatVol * 50;  // preload the battery sum for averaging
  AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
  if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
  if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
  Serial.println("Battery " + String(BatVol) + "\t" + String(float(BatAvg)/BatCal) + "v\t" + String(AnyLong) + "%");
  randomSeed(BatVol); // use the battery voltage as the random seed factor
  // if battery is less than 5.5v then USB mode is assumed, which makes servos inoperable
  if (BatAvg <= 2111) {USB = true; Serial.println(F("USB power assumed."));}
  // Convert BatAvg to a mapped voltage and floating point equivalent
  BatV = getBatV(BatAvg); BatVfp = (float)BatV/10.0;
}

// --------------------------------------------------------------------------------

void Init_ESP_NOW() {
  // called to initiate an ESEP-NOW link or connect to another trannsceiver
  // each time this function is called it will attempt to connect to a different
  // boradcast address, cycling through them, over and over
  //  Serial.println("Init_ESP_NOW()");

  // if we have previously tried a connection then we need to de-initialise it
  if (ESP_NOW_Init) {
    // already attempted to initialise ESP-NOW so strip everything back
    // Serial.println("Closing ESP-NOW");
    esp_now_del_peer(broadcastAddress);
    WiFi.disconnect();
  }
  
  // Init ESP-NOW and returns its status
  if (esp_now_init() != ESP_OK) {
    // Serial.println("Error initializing ESP-NOW");
    return;
  }
  ESP_NOW_Init = true;
  
  // change the broadcast address pointer each time
  ESP_NOW_BA++; if (ESP_NOW_BA > 1) {ESP_NOW_BA = 0;}
  // Serial.println("ESP_NOW_BA = " + String(ESP_NOW_BA));
  
  switch (ESP_NOW_BA) {
    case 0: // Wii Transciever unique MAC: 50:02:91:68:F7:3F
      broadcastAddress[0] = 0x50;
      broadcastAddress[1] = 0x02;
      broadcastAddress[2] = 0x91;
      broadcastAddress[3] = 0x68;
      broadcastAddress[4] = 0xF7;
      broadcastAddress[5] = 0x3F;
      break;
    case 1: // Wii Transciever unique MAC: 50:02:91:68:F0:D2
      broadcastAddress[0] = 0x50;
      broadcastAddress[1] = 0x02;
      broadcastAddress[2] = 0x91;
      broadcastAddress[3] = 0x68;
      broadcastAddress[4] = 0xF0;
      broadcastAddress[5] = 0xD2;
      break;
  }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);

  // Register peer
  esp_now_peer_info_t peerInfo; //initialize and assign the peer information as a pointer to an addres
  memset(&peerInfo, 0, sizeof(peerInfo));
  memcpy(peerInfo.peer_addr, broadcastAddress, 6); //copy the value of  broadcastAddress with 6 bytes to peerInfo.peer_addr
  peerInfo.channel = 0;     //channel at which the esp talk. 0 means undefined and data will be sent on the current channel. 1-14 are valid channels which is the same with the local device 
  peerInfo.encrypt = false; //not encrypted

  //Add the device to the paired device list 
  if (esp_now_add_peer(&peerInfo) != ESP_OK){ 
    Serial.println("Failed to add peer");
    return;
  }

  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
}

// --------------------------------------------------------------------------------

void Isr_En() {
  // Set up a H/W timer to run the ISR interrupt @ 40µs (25kHz)
  IsrRun = false; delayMicroseconds(100);                   // block the ISR code from running
  // hw_timer_t *IsrTimer = NULL;
  StepTimer = timerBegin(IsrTimerNum, IsrPrescale, true);   // start timer IsrTimer (0-3) using 80MHz clock, count up to 40, hence 2MHz
  timerAttachInterrupt(StepTimer, &onTimer, true);          // assign an interrupt to the timer
  timerAlarmWrite(StepTimer, IsrTimeCnt, true);             // set the interrupt to occur at 100µs for 10kHz audio samples
  timerAlarmEnable(StepTimer);                              // enable the ISR
}

// --------------------------------------------------------------------------------

void Isr_1s() {
  // Set up a timer to run the ISR interrupt @ 1 second intervals.
  // This is a very slow clock rate, effectively stopping the ISR function
  IsrRun = false; delayMicroseconds(100);                 // block the ISR code from running
  // hw_timer_t *AudioTimer = NULL;
  StepTimer = timerBegin(IsrTimerNum,IsrPrescale, true);  // start timer IsrTimer (0-3) using 80MHz clock, count up to 40, hence 2MHz
  timerAttachInterrupt(StepTimer, &onTimer, true);        // assign an interrupt to the timer
  timerAlarmWrite(StepTimer, 2000000, true);              // set the interrupt to occur at 65.2µs for 16kHz audio samples
  timerAlarmEnable(StepTimer);                            // enable the ISR
  IsrRun = true;                                            // enable the Isr code
}

// --------------------------------------------------------------------------------

void Motors_OFF() {
  // Disables ISR function, and sets driver enables to OFF
  IsrRun = false;
  Lft_En = HIGH; digitalWrite(StepLftEn,Lft_En);
  Rht_En = HIGH; digitalWrite(StepRhtEn,Rht_En);
  // Motor pulses are set to a long count
  Lft_Motor_Throttle = 1000; Rht_Motor_Throttle = 1000;
}

// --------------------------------------------------------------------------------

void MPU_6050_Initialise() {
  // set up MPU6050 3-axis motion seonsor hardware
  // By default the MPU-6050 sleeps. So we have to wake it up.
  Wire.beginTransmission(MPU_address);  // Start communication with the address found during search.
  Wire.write(0x6B);                     // We want to write to the PWR_MGMT_1 register (6B hex)
  Wire.write(0x00);                     // Set the register bits as 00000000 to activate the gyro
  I2CErr = Wire.endTransmission();      // End the transmission with the MPU.
  if (I2CErr != 0) {return;}
  delayMicroseconds(10);                // Allow MPU time to respond

  //Set the full scale of the gyro to +/- 250 degrees per second
  Wire.beginTransmission(MPU_address);  // Start communication with the address found during search.
  Wire.write(0x1B);                     // We want to write to the GYRO_CONFIG register (1B hex)
  Wire.write(0x00);                     // Set the register bits as 00000000 (250dps full scale)
  I2CErr = Wire.endTransmission();      // End the transmission with the MPU.
  if (I2CErr != 0) {return;}
  delayMicroseconds(10);                // Allow MPU time to respond

  //Set the full scale of the accelerometer to +/- 4g.
  Wire.beginTransmission(MPU_address);  // Start communication with the address found during search.
  Wire.write(0x1C);                     // We want to write to the ACCEL_CONFIG register (1A hex)
  //  Wire.write(0x00);                 // Set the register bits as 00000000 (+/- 2g full scale range)
  Wire.write(0x08);                     // Set the register bits as 00001000 (+/- 4g full scale range)
  I2CErr = Wire.endTransmission();      // End the transmission with the MPU.
  if (I2CErr != 0) {return;}
  delayMicroseconds(10);                // Allow MPU time to respond

  //Set some filtering to improve the raw data.
  Wire.beginTransmission(MPU_address);  // Start communication with the address found during search
  Wire.write(0x1A);                     // We want to write to the CONFIG register (1A hex)
  Wire.write(0x03);                     // Set the register bits as 00000011 (Set Digital Low Pass Filter to ~43Hz)
  I2CErr = Wire.endTransmission();      // End the transmission with the MPU.
  if (I2CErr != 0) {return;}
  delayMicroseconds(10);                // Allow MPU time to respond
}

// --------------------------------------------------------------------------------

void OpBrakeInc(float zV) {
  // Increment/decrement the current Ptch_OpBrake value
  if (DmB == 0) {Ptch_OpBrake += zV;} else {Ptch_OpBrake -= zV;}

  if (Ptch_OpBrake > 99.99) {Ptch_OpBrake = 99.99;}
  else if (Ptch_OpBrake < 0.000) {Ptch_OpBrake = 0.0;}
  DispDel = 1;
}

// --------------------------------------------------------------------------------

void OpModeInc() {
  // Increment/decrement the OpMode flag
       if ((DmB == 0) && (OpMode < 1)) {OpMode++;}
  else if ((DmB == 1) && (OpMode > 0)) {OpMode--;}
}

// --------------------------------------------------------------------------------

void PrintMPU() {
  // send the MPU values to the serial port
  if (PrintDel > 0) {PrintDel--; return;} // wait for buffer to empty

  Serial.print("\nAcX:" + String(AcX));
  Serial.print("\tAcY:" + String(AcY));
  Serial.print("\tAcZ:" + String(AcZ));
  Serial.print("\tGyX:" + String(GyX));
  Serial.print("\tGyY:" + String(GyY));
  Serial.print("\tGyZ:" + String(GyZ));
  PrintDel = 10;
}

// -------------------------------------------------------------------------------

void PtchDinc(float zV) {
  // increment the Pitch PID D-gain by zV
  if (DmB != 0) {
    // mouse right-click
    if (Ptch_d_gain > 0.00) {Ptch_d_gain -= zV;}
    else {Ptch_d_En = false; Ptch_d_gain= 0.00;}
  } else {
         if (!Ptch_d_En) {Ptch_d_En = true; Ptch_d_gain= 0.00;}
    else if (Ptch_d_gain <= 99.99) {Ptch_d_gain += zV;}
  }
  if (Ptch_d_gain < 0.00) {Ptch_d_gain = 0.00;}     // value must be >= 0.0
  if (Ptch_d_gain > 99.99) {Ptch_d_gain = 99.99;} // value must be <= 99.99
}

// -------------------------------------------------------------------------------

void PtchIinc(float zV) {
  // increment the Pitch PID I-gain by zV
  if (DmB != 0) {
    // mouse right-click
    if (Ptch_i_gain > 0.00) {Ptch_i_gain -= zV;}
    else {Ptch_i_En = false; Ptch_i_gain= 0.00;}
  } else {
         if (!Ptch_i_En) {Ptch_i_En = true; Ptch_i_gain= 0.00;}
    else if (Ptch_i_gain <= 99.99) {Ptch_i_gain += zV;}
  }
  // if Ptch_i_gain drops to zero then reset Ptch_i_mem, in case we are self-balancing
  if (Ptch_i_gain < 0.00) {Ptch_i_gain = 0.00; Ptch_i_mem = 0;} // value must be >= 0.0
  if (Ptch_i_gain > 99.99) {Ptch_i_gain = 99.99;}               // value must be <= 99.99
}

// -------------------------------------------------------------------------------

void PtchIMinc(float zV) {
  // increment the Pitch PID I-max by zV
  if (DmB != 0) {if (Ptch_i_max >= 0.00) {Ptch_i_max -= zV;}}
  else {if (Ptch_i_max <= 99.99) {Ptch_i_max += zV;}}
  if (Ptch_i_max < 0.00) {Ptch_i_max = 0.00;}     // value must be >= 0.0
  if (Ptch_i_max > 99.99) {Ptch_i_max = 99.99;} // value must be <= 99.99
}

// -------------------------------------------------------------------------------

void PtchPinc(float zV) {
  // increment the Pitch PID P-gain by zV
  if (DmB != 0) {if (Ptch_p_gain >= 0.00) {Ptch_p_gain -= zV;}}
  else {if (Ptch_p_gain <= 99.99) {Ptch_p_gain += zV;}}
  if (Ptch_p_gain < 0.00) {Ptch_p_gain = 0.00;}     // value must be >= 0.0
  if (Ptch_p_gain > 99.99) {Ptch_p_gain = 99.99;} // value must be <= 99.99
}

// -------------------------------------------------------------------------------

void PtchSinc(float zV) {
  // increment the Pitch PID StartAng by zV
  if (DmB != 0) {if (PtchStartAng >= -9.99) {PtchStartAng -= zV;}}
  else {if (PtchStartAng <= 9.99) {PtchStartAng += zV;}}
  if (PtchStartAng < -9.99) {PtchStartAng = -9.99;}   // value must be >= -9.99
  if (PtchStartAng > 9.99) {PtchStartAng = 9.99;}     // value must be <= 9.99
}

// -------------------------------------------------------------------------------

void PtchSbInc(float zV) {
  // Increment the Pitch SbPtchSpInc by zV.
  // if the function is OFF then turn it on with minimum value
  // if the function is reduced beyond 0.000 then tuirn it OFF
  if (DmB != 0) {
    if (Sb) {
            if (SbSpInc == 0.0000) {Sb = false;}
      else if (SbSpInc >= 0.0000) {SbSpInc -= zV;}
    }
  } else {
         if (!Sb) {Sb = true; SbSpInc= 0.0000;}
    else if (SbSpInc <= 9.9999) {SbSpInc += zV;}
  }
  if (SbSpInc <= 0.0000) {SbSpInc = 0.0000;}      // value must be >= 0.000
  if (SbSpInc >= 9.9999) {SbSpInc = 9.9999;}      // value must be <= 9.999
}

// -------------------------------------------------------------------------------

void readAccelAll() {
  // Read the X,Y,Z-axis acceleration values from the MPU6050
  // Sensor mounting in this robot:
  // AcX = yaw, +ve tip right, -ve tip left
  // AcY = up/down, +ve upside-down, -ve right way up
  // AcZ = pitch, +ve tip forward, -ve tip backward
  t2 = micros();                        // we are timing this I2C function
  AcXL = AcX; AcYL = AcY; AcZL = AcZ;   // track previous values
  Wire.beginTransmission(MPU_address);
  Wire.write(0x3B);                     // starting with register 0x3B (ACCEL_XOUT_H)
  I2CErr = Wire.endTransmission();      // End the transmission with the MPU.
  if (I2CErr != 0) {return;}            // abort on errors
  delayMicroseconds(I2Cdel);                // Allow MPU time to respond

  Wire.requestFrom(MPU_address,6);      // request 6 register read
  AcX = Wire.read()<<8|Wire.read();     // 0x3B (ACCEL_XOUT_H) & 0x3C (ACCEL_XOUT_L)
  AcY = Wire.read()<<8|Wire.read();     // 0x3D (ACCEL_YOUT_H) & 0x3E (ACCEL_YOUT_L)
  AcZ = Wire.read()<<8|Wire.read();     // 0x3F (ACCEL_ZOUT_H) & 0x40 (ACCEL_ZOUT_L)

  // Serial.print(AcX); Serial.print("\t");
  // Serial.print(AcY); Serial.print("\t");
  // Serial.println(AcZ);
  
  // apply calibration offsets to these raw values
  AcX -= AccOffX;  // note AcXOff should be zero initially in order to determin its value
  AcY -= AccOffY;  // note AcYOff should be zero initially in order to determin its value
  AcZ -= AccOffZ;  // note AcZOff should be zero initially in order to determin its value

  // test for rapid Pitch accelerometer overshoot, due to vibration
  // setting AccRampY too low will impact on angle calculations
  if (AccRampX > 0) {
    AccDif = AcX - AcXL; if (fabs(AccDif) > AccRampX) {if (AcX > AcXL) {AcX = AcXL + AccRampX;} else {AcX = AcXL - AccRampX;}}
  }
  if (AccRampY > 0) {
    AccDif = AcY - AcYL; if (fabs(AccDif) > AccRampY) {if (AcY > AcYL) {AcY = AcYL + AccRampY;} else {AcY = AcYL - AccRampY;}}
  }
  if (AccRampZ > 0) {
    AccDif = AcZ - AcZL; if (fabs(AccDif) > AccRampZ) {if (AcZ > AcZL) {AcZ = AcZL + AccRampZ;} else {AcZ = AcZL - AccRampZ;}}
  }

  //##############################################
  // Averaging filters to reduce MPU random noise
  //##############################################
  // set AvCnt > 1 to increase averaging effect, default AvCnt = 1
  // note this form of filtering introduces a phase delay, which may lead to instability
  if (AvCnt > 1) {
    AcXSum = AcXSum - AcXAvg + AcX; AcXAvg = AcXSum /AvCnt; AcX = AcXAvg;
    AcYSum = AcYSum - AcYAvg + AcY; AcYAvg = AcYSum /AvCnt; AcY = AcYAvg;
    AcZSum = AcZSum - AcZAvg + AcZ; AcZAvg = AcZSum /AvCnt; AcZ = AcZAvg;
  }

  // limit values to avoid trig function problems
  if (AcX > 8192) {AcX = 8192;} else if (AcX < -8192) {AcX = -8192;}
  if (AcY > 8192) {AcY = 8192;} else if (AcY < -8192) {AcY = -8192;}
  if (AcZ > 8192) {AcZ = 8192;} else if (AcZ < -8192) {AcZ = -8192;}
  t3 = micros() - t2;                   // capture time spent in t3, 980us
}

// -------------------------------------------------------------------------------

void readGyrosAll() {
  // read the X,Y,Z-axis gyroscope values from the MPU
  // note gyro values are in degrees/sec so we use time between measurements to 
  // determine the average displaced angle
  //  t0 = micros();  // t1-t0 average = 315µs, longer with I2C bus contention
  GyXL = GyX; GyYL = GyY; GyZL = GyZ;   // track previous values
  Wire.beginTransmission(MPU_address);
  Wire.write(0x43);                     // starting with register 0x43 (GYRO_XOUT_H)
  I2CErr = Wire.endTransmission();      // End the transmission with the MPU.
  if (I2CErr != 0) {return;}            // abort on errors
  delayMicroseconds(I2Cdel);            // Allow MPU time to respond

  Wire.requestFrom(MPU_address,6);      // request 6 register reads
  GyX = Wire.read()<<8|Wire.read();     // 0x43 (GYRO_XOUT_H) & 0x44 (GYRO_XOUT_L)
  GyY = Wire.read()<<8|Wire.read();     // 0x45 (GYRO_YOUT_H) & 0x46 (GYRO_YOUT_L)
  GyZ = Wire.read()<<8|Wire.read();     // 0x47 (GYRO_ZOUT_H) & 0x48 (GYRO_ZOUT_L)

  // Serial.print(GyX); Serial.print("\t");
  // Serial.print(GyY); Serial.print("\t");
  // Serial.println(GyZ);

  if (GyT > 0) {
    GyTD = micros() - GyT;  // get the differential time between readings
    GyT = GyTD + GyT;       // set the timer for the next reading
  } else {
    // 1st reading after reset so set nominal 5ms
    GyT = micros();
    GyTD = 5000; GyXL = GyX; GyYL = GyY; GyZL = GyZ;
  }

  // apply calibration offsets to these raw values
  GyX -= GyrOffX;  // note GyrOffX should be zero initially in order to determin its value
  GyY -= GyrOffY;  // note GyrOffY should be zero initially in order to determin its value
  GyZ -= GyrOffZ;  // note GyrOffZ should be zero initially in order to determin its value

  //##############################################
  // Averaging filters to reduce MPU random noise
  //##############################################
  GyXSum = GyXSum - GyXAvg + GyX; GyXAvg = GyXSum /AvCnt; GyX = GyXAvg;
  GyYSum = GyYSum - GyYAvg + GyY; GyYAvg = GyYSum /AvCnt; GyY = GyYAvg;
  GyZSum = GyZSum - GyZAvg + GyZ; GyZAvg = GyZSum /AvCnt; GyZ = GyZAvg;
  //  t1 = micros();
}

// --------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set PrintTgt to indicate source as serial port for responses
  keyVal = Serial.read();
  if (keyVal != -1) {SerialRx = true; PrintTgt = 0; decodeKey(keyVal);}
}

// ---------------------------------------------------------------------------------

void readSW0() {
  // Read the right button switch and respond accordingly, pressed == LOW
  // A button press will automatical drop the current task
  // A single press will toggle the laser range finder
  // A double press will enter Back Away mode
  // If pressed (held) during boot the WiFRi will enter BLE mode
  // SW0 is read every 20ms (50Hz) to debounce
  sw0State = digitalRead(sw0Pin); // record button state
  if (sw0_Nop) {return;}          // block this function whilst == true
  
  if (sw0State == LOW) {
    //##############################################################################
    //
    // SW0 button is pressed down
    //
    //##############################################################################
    if (sw0Wup) {return;}
    
    if (sw0LastState == HIGH) {
      // SW0 has just been pressed down
      //  Serial.println("SW0 Hi-LO");
      if (TEST) {Serial.println("SW0 Hi-LO");}
      sw0DwnTime = 0; // restart button down time counter
      sw0Cnt++;       // count on the falling edge
      sw0Timer = 0;   // reset the timer
      //  if (WiFiEn) {SetMainMode(0); return;}
    } else {
      // whilst the button is down adjust the timer at 20ms steps
      // if Wii is not connected test for a long press to invoke demo mode
      sw0DwnTime++; // track button pressed time
      if (TEST) {
        // look for long press > 1 sec in TEST mode
        if (sw0Timer == 120) {
          // 2 second timeout
          sw0Cnt = 0; sw0Timer = 0;
        }
      } else {
        if (sw0Timer > 30) {
          // button SW0 held down for > 0.5 second
          // block further actions until released
          sw0Cnt = 0; sw0Timer = 0; sw0Wup = true;
        }
      }
    }
  } else {
    //##############################################################################
    //
    // SW0 button is released
    //
    //##############################################################################
    if (sw0Wup) {
      // waited for button release
      sw0Wup = false; sw0Cnt = 0; sw0Timer = 0;
    } else {
      if (sw0LastState == LOW) {
        // SW0 has just been released
        //  Serial.println("SW0 Lo-Hi");
        if (TEST) {Serial.println("SW0 Lo-Hi");}

      }
      if (sw0Timer > 50) {
        // button released for 1 sec so assume valid button counts
         Serial.print(F("sw0Cnt = ")); Serial.println(sw0Cnt);
        if (TEST && (sw0Cnt > 0)) {
          // single button press during test mode
        }
        if (sw0Cnt == 1) {
          // Toggle the LTOF enabled state.
          if (RangeEn) {
            VL53L1X_OFF(); Display_Text_2x16("Ranging","Disabled",25); DispMode = 1;
          } else {
            VL53L1X_ON(); Display_Text_2x16("Ranging","Enabled",25);
            if (MainMode) {DispMode = 11;} else {DispMode = 10;}
          }
          BackAwayTask = 0; // Cancel Back Away functions.
        }
        if (sw0Cnt == 2) {
          VL53L1X_ON(); Display_Text_2x16("Back Away","Active",25);
          BackAwayTask = 1; // Initialise Back away functions
        }
        sw0Cnt = 0; sw0Timer = 0;
      }
      sw0DwnTime = 0;
    }
  }
  sw0LastState = sw0State; // record current state
  if (sw0Cnt > 0) {sw0Timer++;}
}

// ---------------------------------------------------------------------------------

void readSW1() {
  // Read the left button switch and respond accordingly, pressed == LOW
  // A button press will automatical drop the self-balancing mode.
  // If pressed (held) during boot the WiFRi will enter BLE mode
  // SW1 is read every 20ms (50Hz) to debounce
  sw1State = digitalRead(sw1Pin); // record button state
  if (sw1_Nop) {return;}          // block this function whilst == true
  
  if (sw1State == LOW) {
    //##############################################################################
    //
    // SW1 button is pressed down
    //
    //##############################################################################
    if (sw1Wup) {return;}
    
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
  //      Serial.println("SW1 Hi-LO");
      if (TEST) {Serial.println("SW1 Hi-LO");}
      sw1DwnTime = 0; // restart button down time counter
      sw1Cnt++; // count on the falling edge
      sw1Timer = 0; // reset the timer
    } else {
      sw1DwnTime++; // track button pressed time
      if(sw1Timer >= 60) {
        // button held down for >= 1 second
        sw1Cnt = 0; sw1Timer = 0;
      }
    }
    if (SafeMode > 0) {SafeMode = -1;}  // reset SafeMode
  } else {
    //##############################################################################
    //
    // sw1State == HIGH as SW1 button is not pressed
    //
    //##############################################################################
    if (sw1Wup) {
      // release locked-out state
      sw1Wup = false;
      sw1Cnt = 0; sw1Timer = 0;
    } else {
      // normal SW1 release
      if (sw1LastState == LOW) {
        // SW1 has just been released
  //        Serial.println("SW1 Lo-Hi");
        if (TEST) {Serial.println("SW1 Lo-Hi");}
      }
      if (!TEST) {
        // not in TEST mode so set task based on count value
          sw1Cnt = 0; sw1Timer = 0; // clear events
      }
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// ---------------------------------------------------------------------------------

void readWii() {
  // Called every 20ms when WiFi is connected, to check if data has
  // been received from the Wii controller.
  if (BackAwayTask > 0) {return;} // Block this function when in Back Away mode

  if (RxRec) {
    // Valid data has been received, as a single bite.
    // This robot does not use proportional X/Y joystick commands, only direct
    RxRec = false; WiFiRxCnt = 25;  // auto-cancel controller demands after 1 second
    RxByte = 0b00000000;            // clear the byte mask
    // construct RxByte from received data, for use in PID controller
    // 00000000
    // ||||||||
    // |||||||!--- B0 turn right bit
    // ||||||!---- B1 turn left bit
    // |||||!----- B2 go forward bit
    // ||||!------ B3 go backwards bit
    // |||!------- B4 Z button
    // ||!-------- B5 C button
    // |!--------- B6 not used
    // !---------- B7 not used
    // do this when in self-balance or in TEST mode
    if ((MainMode == 1) || (SafeMode == 4)) {
           if (JoyX >= 132) {RxByte  = 0b00000001;}
      else if (JoyX <= 124) {RxByte  = 0b00000010;}
           if (JoyY >= 132) {RxByte |= 0b00000100;}
      else if (JoyY <= 124) {RxByte |= 0b00001000;}
      if ((CZ & 0b01) == 0) {RxByte |= 0b00010000;}
      if ((CZ & 0b10) == 0) {RxByte |= 0b00100000;}
    }
  }
}

// ---------------------------------------------------------------------------------

void readWiFiRx() {
  // Called from the main loop every cycle to check for ESP-NOW data.
  // A state machine is used to separate the data blocks by header type.
  switch(Rx_Task) {
    case 0: // determine the type of data to handle
           if (Rx_Buff.ESPdata[Rx_Pnt] == 'B') {Rx_Task = 3; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'C') {WiiType = 'C'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'N') {WiiType = 'N'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '-') {WiiType = '-'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '$') {Rx_Task = 2; break;}
      // oh dear! block header not recognised so discard the block and flag error
      else {WiFiRxErr++; WiFiRx = false;}
      break;
      
    case 1: // load Wii I2C data
      // After the header byte there are 6 data bytes + checksum
      // Perform XOR on 6 bytes to check for errors against checksum
      RxClass = false;  // clear the 'Classic' flag
      RxNunch = false;  // clear the 'Nunchuk' flag
      if (Rx_Buff.ESPdata[7] == (Rx_Buff.ESPdata[1] ^ Rx_Buff.ESPdata[2] ^ Rx_Buff.ESPdata[3] ^ Rx_Buff.ESPdata[4] ^ Rx_Buff.ESPdata[5] ^ Rx_Buff.ESPdata[6])) {
        // data has passed the checksum test
        RxWiFi[0] = Rx_Buff.ESPdata[1];
        RxWiFi[1] = Rx_Buff.ESPdata[2];
        RxWiFi[2] = Rx_Buff.ESPdata[3];
        RxWiFi[3] = Rx_Buff.ESPdata[4];
        RxWiFi[4] = Rx_Buff.ESPdata[5];
        RxWiFi[5] = Rx_Buff.ESPdata[6];

        RxRec = true;     // indicate a valid frame
        if (WiiType == 'C') {
          // if Wii Classic data strip off the JLX,JLY,JRX,JRY and CZ values
          // the left joystick is used for head/gun control
          // the right joystick is used for track drive control
          // Nunchuk 'C' and 'Z' buttons are mapped to BRT and BLT respectively here
          // the right hand X axis is 5 bits, 0 - 31, centred on 16
          // here the bits are extracted and shifted, to centre X on 128
          JoyX = (RxWiFi[0] & 0b11000000) + ((RxWiFi[1] & 0b11000000)>>2) + ((RxWiFi[2] & 0b10000000)>>4);   // joystick X 0-248
          if (JoyX > 128) {JoyX = map(JoyX,128,248,128,255);} // remap to 128 - 255
          // the right hand Y axis is 5 bits, 0 - 31, centred on 16
          // here the bits are extracted and shifted, to centre Y on 128
          JoyY = (RxWiFi[2] & 0b00011111)<<3;                 // joystick Y 0-248
          if (JoyY > 128) {JoyY = map(JoyY,128,248,128,255);} // remap to 128 - 255
          // Serial.println("X=" + String(JoyX) + "\tY=" + String(JoyY));
          CZ = (RxWiFi[4] & 0b00000010) + ((RxWiFi[4] & 0b00100000)>>5); // CZ buttons, C = 2/0, Z = 1/0, active low
          RxClass = true;
        } else if (WiiType == 'N') {
          // if Wii Nunchuk data strip off the JX,JY and CZ values
          JoyX = RxWiFi[0];   // joystick X 0-255
          JoyY = RxWiFi[1];   // joystick Y 0-255
          CZ = RxWiFi[5] & 3; // CZ buttons, C = 2/0, Z = 1/0, active low
          RxNunch = true;
        } else {
          // no Nunchuk so set defaults
          JoyX = 128; JoyY = 128; CZ = 3;
        }
      } else {
        // received data failed the checksum test
        WiFiRxErr++; RxRec = false;
      }
      Rx_Task = 0;  // reset the task pointer to lok at another block if needed
      Rx_Pnt += 8;  // increase the data pointer beyond this block
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0;} // reached end of buffer
      break;

    case 2: // load text data as if from a serial port
      // read characters up to the end of the buffer
      Rx_Pnt++;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0; break;} // reached end of buffer
      keyVal = Rx_Buff.ESPdata[Rx_Pnt]; PrintTgt = 1; decodeKey(keyVal);
  //      Serial.write(keyVal);
      break;

    case 3: // load binary data block
      // data length is specified in second byte
      Rx_Pnt++;
      // at the moment just jump over it, binary transfer is not needed
      Rx_Pnt += Rx_Buff.ESPdata[Rx_Pnt] + 1;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; break;} // reached end of buffer
      break;
  }
}

// ---------------------------------------------------------------------------------

void RxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Rx buffer of length zlen
  Rx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Rx_Chk = Rx_Chk ^ Rx_Buff.ESPdata[zP];
  }
}

// --------------------------------------------------------------------------------

void set_LED0_New(int16_t zT) {
  // Called by Core 1 to trigger Core 0 to load a new task.
  LED0_New = zT; LED0_Set = true;
}

// --------------------------------------------------------------------------------

void set_LED1_New(int16_t zT) {
  // Called by Core 1 to trigger Core 0 to load a new task.
  LED1_New = zT; LED1_Set = true;
}

// --------------------------------------------------------------------------------

void set_LED0_Task(int16_t zT) {
  // This starts a new LED0 task, it must be called from Core 0.
  // If zT matches the current task then do nothing.
  // If zT is negative then force the reset process.
  if (LED0_Task == zT) {return;}

  Bright0 = 255;        // FastLED 0 brightness
  LED0_Cnt = 0;         // general counter used in LED tasks
  LED0_Del = 0;         // LED Task delay, adds pauses into the processes
  LED0_Flg = false;     // LED flag used in simple flashing mode
  LED0_Inc = 1;         // LED Task increment variable
  LED0_Pnt = 0;         // LED pointer variable
  LED0_Run = true;      // enable LED tasks
  LED0_Set = false;     // flag used to initiate Gun LED task change
  LED0_SubTsk = 0;      // LED sub-task pointer
  if (zT != LED0_Task) {LED0_Last = LED0_Task;} // record current task before switching

  LED0_Task = zT;  // LED task pointer
}

// --------------------------------------------------------------------------------

void set_LED1_Task(int16_t zT) {
  // This starts a new LED1 task, it must be called from Core 0.
  // If zT matches the current task then do nothing.
  // If zT is negative then force the reset process.
  if (LED1_Task == zT) {return;}

  Bright1 = 255;        // FastLED 1 brightness
  LED1_Cnt = 0;         // general counter used in LED tasks
  LED1_Del = 0;         // LED Task delay, adds pauses into the processes
  LED1_Flg = false;     // LED flag used in simple flashing mode
  LED1_Inc = 1;         // LED Task increment variable
  LED1_Pnt = 0;         // LED pointer variable
  LED1_Run = true;      // enable LED tasks
  LED1_Set = false;     // flag used to initiate Gun LED task change
  LED1_SubTsk = 0;      // LED sub-task pointer
  if (zT != LED1_Task) {LED1_Last = LED1_Task;} // record current task before switching

  LED1_Task = zT;  // LED task pointer
}

// ---------------------------------------------------------------------------------

void Set_M210() {
  // called to set the driver M pins to match M210
  if ((M210 & 0b100) > 0) {digitalWrite(M2_Pin,HIGH);} else {digitalWrite(M2_Pin,LOW);}
  if ((M210 & 0b010) > 0) {digitalWrite(M1_Pin,HIGH);} else {digitalWrite(M1_Pin,LOW);}
  if ((M210 & 0b001) > 0) {digitalWrite(M0_Pin,HIGH);} else {digitalWrite(M0_Pin,LOW);}
}

// ---------------------------------------------------------------------------------

void Set_Speeds(int16_t zSpd) {
  // sets the speeds for driving and turning, based on 3 settings
  switch (zSpd) {
    case 1:
     Drive_Max  = 40.0; PtchSpTgt = 1.9;
     TurnSpdMax = 20.0; TurnSpdMin =  4.0; TurnSpdInc = 0.1; break;
    case 2:
     Drive_Max = 60.0; PtchSpTgt = 2.1;
     TurnSpdMax = 24.0; TurnSpdMin = 10.0; TurnSpdInc = 0.1; break;
    case 3:
     Drive_Max = 80.0; PtchSpTgt = 2.3;
     TurnSpdMax = 28.0; TurnSpdMin = 16.0; TurnSpdInc = 0.1; break;
    case 4:
     Drive_Max = 100.0; PtchSpTgt = 2.5;
     TurnSpdMax = 34.0; TurnSpdMin = 22.0; TurnSpdInc = 0.1; break;
  }
  Gear = zSpd; // record current setting
}

// ---------------------------------------------------------------------------------

void SpTgtInc(float zV) {
  // increment/decrement the current PtchSpTgt value
  if (DmB == 0) {PtchSpTgt += zV;} else {PtchSpTgt -= zV;}

  if (PtchSpTgt > 9.90) {PtchSpTgt = 9.90;}
  else if (PtchSpTgt < 0.000) {PtchSpTgt = 0.000;}
  DispDel = 1;
}

// ---------------------------------------------------------------------------------

void SpdTgtInc(float zV) {
  // increment/decrement the current TgtSpdMax value
  if (DmB == 0) {Drive_Max += zV;} else {Drive_Max -= zV;}

  if (Drive_Max > 300.0) {Drive_Max = 300.0;}
  else if (Drive_Max < 1.000) {Drive_Max = 1.0;}
  DispDel = 1;
}

// ---------------------------------------------------------------------------------

void ToneAttach(int zFreq) {
  // attaches a tone channel for playing notes using ledcWriteNote()
  ledcSetup(PwmChTone,zFreq,12);
  ledcAttachPin(TonePin,PwmChTone);
  ToneAtt = true;
}

// ---------------------------------------------------------------------------------

void ToneDetach() {
  // detaches the tone channel & set the output pin LOW
  ledcDetachPin(TonePin);
  pinMode(TonePin,OUTPUT); digitalWrite(TonePin,LOW);
  ToneAtt = false; ToneFreq = 0;
}

// ---------------------------------------------------------------------------------

void ToneNote(note_t zNote,uint8_t zOct) {
  // play a note using the tone library
  if (!ToneAtt) {ToneAttach(1);}  // ensure PWM channel is assigned
  ledcWriteNote(PwmChTone, zNote, zOct);
  NoteLast = zNote;
}

// ---------------------------------------------------------------------------------

void TonePlay(int zFreq) {
  // play a note on the tone channel
  // if not attached, then attach it
  if (!ToneAtt) {ToneAttach(zFreq);}
  ledcWriteTone(PwmChTone,zFreq);
  ToneFreq = zFreq; // record tone value for sweeping later
}

// --------------------------------------------------------------------------------

void VL53L1X_OFF() {
  // switches OFF the VL53L1X laser ranging device
  if (!VL53L1X_Task) {return;}  // not running

  digitalWrite(XSHUT_PIN,LOW);  // shut down VL53L1X sensor
  // Serial.println("VL53L1X OFF");
  RangeEn = false; VL53L1X_Task = 0; LTOFfps = 0;
}

// --------------------------------------------------------------------------------

void VL53L1X_ON() {
  // Switches ON the VL53L1X laser ranging device using XSHUT pin state.
  // If it was already ON then switch it OFF, wait, then ON again.
  if (RangeEn) {VL53L1X_OFF(); delay(1);}
  // synchLoopTimers(1);}
  
  digitalWrite(XSHUT_PIN,HIGH);     // enable VL53L1X sensor
  if (VL53L1X.init() == false) {
    // ***WARNING*** you can't allow the audio ISR to run whilst using I2C
    // so here we temporarily disable it, to prevent the system from crashing.
    // We store the IsrRun flag, and this must be replaced before leaving this function.
    // So if you add to this code, ensure that you respect that process.
    // Serial.println("VL53L1X sensor detected!");
    I2C_LTOF = true;
    IsrRunF = IsrRun; IsrRun = false; // temporarily block the audio ISR
    VL53L1X.setDistanceModeShort();         // Sets to 1.3M range
    VL53L1X.setTimingBudgetInMs(33);        // Set timing budget to 33 ms
    VL53L1X.setIntermeasurementPeriod(33);  // Set measurement period in ms
    VL53L1X.setROI(VL53L1X_ROIX,VL53L1X_ROIY,VL53L1X_OC); // Set SPAD ROI
    VL53L1X.startRanging();                 // Starts taking measurements
    // Serial.println("VL53L1X ROI SPAD W= " + String(VL53L1X.getROIX()));
    // Serial.println("VL53L1X ROI SPAD H= " + String(VL53L1X.getROIY()));
    RangeEn = true; VL53L1X_Task = 1; VL53L1X_Skip = millis();
    Range = LtofLimit; RangeLast = LtofLimit;
    IsrRun = IsrRunF; // return IsrRun to its previous value
  } else {
    I2C_LTOF = false; RangeEn = false;
    // Serial.println("VL53L1X sensor failed!");
  }
}

// --------------------------------------------------------------------------------

void VL53L1X_Run() {
  // Control the VL53L1X sensor using a series of tasks, if task > 0
  // This code gets timing information, for using the sensor more efficiently
  // Once measurements start, a skip timer is used to avoid unnecessary I2C readings
  if (millis() < VL53L1X_Skip) {return;}  // allowing for reading delay 32ms
  

  // ***WARNING*** you can't allow the audio ISR to run whilst using I2C
  // so here we temporarily disable it, to prevent the system from crashing.
  // We store the IsrRun flag, and this must be replaced before leaving this function.
  // So if you add to this code, ensure that you respect that process.
  IsrRunF = IsrRun; IsrRun = false; // temporarily block the ISR
  switch(VL53L1X_Task) {
    // case 8 is placed at the front of the list for speed
    case 8: // wait for measurement to complete
      // This case is placed at the head of the list as it will be used most often
      // If MPU6050 data is not ready we keep coming to this task
      // We save previous range. just before taking a new measurement
      // ***NOTE*** I had a faulty sensor, which produced noise rather than a long range value.
      // Therefore this code includes a filter which tries to make sense of the readings,
      // and reject any noisy values as if out of range.
      if (VL53L1X.checkForDataReady()) {
      //  t1 = millis() - t1; // how long did the measurement take?
      //  Serial.println("VL53L1X.checkForDataReady() in " + String(t1) + "ms");
      //  t1 = millis();
        VL53L1X_Skip = millis() + VL53L1X_SkipDel; // set the next measurement delay
        VL53L1X_Status = VL53L1X.getRangeStatus();
        RangeLast = RangeRaw;
      //  Serial.println(VL53L1X_Status);
        if (VL53L1X_Status == 0) {
          // no ranging errors, valid measurement
        //  t0 = micros();
          RangeRaw = VL53L1X.getDistance();
        //  t0 = micros() - t0;
        //  Serial.println("VL53L1X.getDistance() in " + String(t0) + "µs");
        } else {
          // if an error occurs we look at the error
          //  1 - Sigma estimator, uncertain measurement, above the internal threshold
          //  2 - Signal value is below the internal defined threshold
          //  4 - Raised when phase is out of bounds
          //  5 - Raised in case of HW or VCSEL failure
          //  7 - Wrapped target, not matching phases
          //  8 - Internal algorithm underflow or overflow
          // 14 -  The reported range is invalid
          // we reject a range if the error exceeds 4
          // if (VL53L1X_Status < 5) {
          //   RangeRaw = VL53L1X.getDistance();
          // } else {RangeRaw = RangeMax;}
          RangeRaw = RangeMax;
        }
        LTOFt1 = millis();
      //  Serial.println(RangeRaw);
      //  t0 = micros();
        VL53L1X.clearInterrupt(); // 444µs clears the interrupt and data ready flag
      //  t0 = micros() - t0;
      //  Serial.println("VL53L1X.clearInterrupt() in " + String(t0) + "µs");
      //  Serial.println("");
        RangeNew = true;
        // adjust the high/low trackers
        if (RangeRaw < LtofLimit) {
          if (RangeTrkHi <= RangeRaw) {
            RangeTrkHi = RangeRaw;
            if (RangeTrkLo < (RangeTrkHi - RangePkPk)) {RangeTrkLo = RangeTrkHi - RangePkPk;}
          } else {RangeTrkHi--;}
          if (RangeTrkLo >= RangeRaw) {
            RangeTrkLo = RangeRaw;
            if (RangeTrkHi > (RangeTrkLo + RangePkPk)) {RangeTrkHi = RangeTrkLo + RangePkPk;}
          } else {RangeTrkLo++;}
        }
        if (abs(RangeLast - RangeRaw) >= RangePkPk) {
          // This is a large change, which Looks like noise
          if (RangeCnt < RangeCntMax) {RangeCnt+= 3;}
        } else if (RangeCnt > -RangeCntMax) {RangeCnt--;}
        if (RangeCnt > 0) {Range = LtofLimit; RangeRawF = Range;}
        else {
          // Sensor reading RangeRaw is rate limited to remove transient noise.
          // if change is less then Range is averaged over two measurements to reduce noise.
          if (RangeRaw > RangeRawF) {
            if ((RangeRaw - RangeRawF) > RangeRate){RangeRawF+= RangeRate;} else {RangeRawF = (RangeRaw + RangeRawF)/2;}
          } else if (RangeRaw < RangeRawF) {
            if ((RangeRawF - RangeRaw) > RangeRate){RangeRawF-= RangeRate;} else {RangeRawF = (RangeRaw + RangeRawF)/2;}
          } Range = RangeRawF;
        }
        if (RangeRaw <= RangeMin) {RangeCnt = RangeCntMax; Range = LtofLimit; RangeRawF = Range;}
      //  Serial.println(String(Range) + ",0,1000"); // print range for serial monitor/plotter
        // limit the range to avoid erroneous values
        if (Range >= LtofLimit) {Range = LtofLimit; RangeUL = true;} else {RangeUL = false;}
        if (Range <= LtofMin) {Range = LtofMin; RangeLL = true;} else {RangeLL = false;}
        // Serial.println(String(RangeRawF) + "," + String(Range) + ",0,1000"); // print range for serial monitor/plotter
      //  Serial.println("Range = " + String(Range) + "mm");
      //  Serial.println(String(Range) + "," + String(servoVal[0]) + "," + String(SteerAng));
        // Store the range at the current angle 30 - 150 == 0 - 40 in RangeData[]
         // calculate the intermeasurement period
        LTOFperiod = LTOFt1 - LTOFt0; LTOFt0 = LTOFt1;
        // calculate the fps frequency
        LTOFfps = 1000/LTOFperiod;
        VL53L1X_Task = 8; // look for another measurement
        if (LTOFcalc) {VL53L1X_Task = 9;}
       else {
        //  Serial.println("0,330," + String(RangeRawF) + "," + String(Range));
        }
      } break;
    case 9: // speed calculation is enabled
      // due to noise on the sensor we cannot measure low speed values, and therefore
      // set s threshold of 10mm which is equivalent to 300mm/s minimum
      RangeDiff = abs(RangeRaw - RangeLast); // get the difference
      if (RangeDiff > 10) {
        // movement is > 10mm so calculate speed
        // we apply a peak detector, with 1 sec sustain window
        LTOFspd10 = (RangeDiff * 10)/LTOFperiod;
        if (LTOFspd10 > LTOFspd10L) {LTOFspd10L = LTOFspd10; LTOFspdCnt = 30;}
        else {LTOFspd10 = LTOFspd10L;}
      }
      if (LTOFspdCnt > 0) {
        LTOFspdCnt--;
        if (LTOFspdCnt < 1) {LTOFspd10 = 0; LTOFspd10L = 0;}
      }
    //  Serial.println("0,1000," + String(Range) + "," + String(LTOFspd10));
      VL53L1X_Task = 8; // look for another measurement
      break;

    case 1: // switch on sensor
      digitalWrite(XSHUT_PIN,HIGH);  // enable VL53L1X sensor
      VL53L1X_Timer = micros(); 
    //  Serial.println("VL53L1X XSHUT HIGH");
      VL53L1X_Task++; break;
    case 2: // check for boot, will time-out after 10ms if sensor does not respond
    //  t0 = micros();
      if (VL53L1X.checkBootState()) {
      //  t0 = micros() - t0;
        VL53L1X_Timer = micros() - VL53L1X_Timer;
      //  Serial.println("VL53L1X.checkBootState() in " + String(t0) + "µs");
        // 825µs needed to boot the device from XSHUT going high
      //  Serial.println("VL53L1X Booted! in " + String(VL53L1X_Timer) + "µs");
        VL53L1X_Task++;
      }
      else if (micros() >= (VL53L1X_Timer + 10000)) {VL53L1X_Task = 99;} // time-out
      break;
    case 3: // device booted, so initialise
    //  t0 = micros();
      if (VL53L1X.init() == false) {
      //  t0 = micros() - t0;
      //  Serial.println("VL53L1X.init() in " + String(t0) + "µs"); // 151ms
        VL53L1X_Task++;
      } else {VL53L1X_Task = 99;} // failed to initialise, so turn sensor OFF
      break;
    case 4: // set the distance mode
    //  t0 = micros();
      VL53L1X.setDistanceModeShort();         // Sets to 1.3M range
    //  t0 = micros() - t0;
    //  Serial.println("VL53L1X.setDistanceModeShort() in " + String(t0) + "µs");
      VL53L1X_Task++; break;
    case 5: // set the timing budget as 33ms
    //  t0 = micros();
      VL53L1X.setTimingBudgetInMs(33);        // Set timing budget to 33 ms
    //  t0 = micros() - t0;
    //  Serial.println("VL53L1X.setTimingBudgetInMs(33) in " + String(t0) + "µs");
      VL53L1X_Task++; break;
    case 6: // set the distance mode
    //  t0 = micros();
      VL53L1X.setIntermeasurementPeriod(33);  // Set measurement period in ms
    //  t0 = micros() - t0;
    //  Serial.println("VL53L1X.setIntermeasurementPeriod(33) in " + String(t0) + "µs");
      RangeEn = true; VL53L1X_Task++; break;
    case 7: // initiate ranging
    //  t0 = micros();
      VL53L1X.startRanging();                 // 446µs Starts taking measurements
    //  t0 = micros() - t0;
    //  Serial.println("VL53L1X.startRanging() in " + String(t0) + "µs");
    //  t1 = millis();
      VL53L1X_Skip = millis() + VL53L1X_SkipDel;  // set the initial delay
      LTOFt0 = millis(); VL53L1X_Task++;          // task 8 is earlier in this code
      RangeCnt = 0;                               // clear the discriminator
      break;
      
    case 99: // a time-out has occured or we want to turn the sensor OFF
      digitalWrite(XSHUT_PIN,LOW);  // disable VL53L1X sensor
    //  Serial.println("VL53L1X XSHUT LOW");
      RangeEn = false; VL53L1X_Task = 0; break;
  }
  IsrRun = IsrRunF;
}

// ---------------------------------------------------------------------------------



  //###############################################################################
  //
  //  WiFi Code
  //
  //###############################################################################

void WiFiClearRxBuff(int zNum) {
  // clears the contents of the receive buffer
  for (int zP = 0; zP < zNum;zP++) {Rx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiClearTxBuff(int zNum) {
  // clears the contents of the transmit buffer
  for (int zP = 0; zP < zNum;zP++) {Tx_Buff.ESPdata[zP] = 0;}
}

// ---------------------------------------------------------------------------------

void WiFiDisable() {
  // when either a switch is pressed or WiFi disable is required
  WiFiEn = false; WiFiCntC = 0; WiFiRestCnt = 0; C_Dn = 8; Z_Dn = 8;
}

// --------------------------------------------------------------------------------

void WiFiDisconnect() {
  // enter the WiFi disconnected state
  WiFiConnected = false;
  WiFiClearRxBuff(WiFiTx_len);
  WiFiClearTxBuff(WiFiTx_len);
  WiFiTryCnt = 0;     // reset the connection retry counter
  WiFiTx_CB = 1;      // set ready to send flag
  WiFiRxRec = false;  // clear the data received flag
  // Serial.println("WiFiDisconnected!");
}

// --------------------------------------------------------------------------------

void WiFiTryToConnect() {
  // try to connect with the Wii Transceiver by sending devices name
  // initialise ESP-NOW link first
  Init_ESP_NOW();

  Tx_Buff.ESPdata[ 0] = 'B';
  Tx_Buff.ESPdata[ 1] = 'a';
  Tx_Buff.ESPdata[ 2] = 'l';
  Tx_Buff.ESPdata[ 3] = 'B';
  Tx_Buff.ESPdata[ 4] = 'o';
  Tx_Buff.ESPdata[ 5] = 't';
  Tx_Buff.ESPdata[ 6] = ' ';
  Tx_Buff.ESPdata[ 7] = 'B';
  Tx_Buff.ESPdata[ 8] = 'L';
  Tx_Buff.ESPdata[ 9] = 'E';
  WiFiTx_len = 11;
  //  t0 = micros();
  esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
  if (result == ESP_OK) {
  //    Serial.println("Sent with success");
  } else {
  //    Serial.println("Error sending the data");
  } 

  WiFiTryNum++; // count the total number of attempts
  //  Serial.println("WiFiTryToConnect...");
}

// --------------------------------------------------------------------------------

void WiFiTxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Tx buffer of length zlen
  WiFiTx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    WiFiTx_Chk = WiFiTx_Chk ^ Tx_Buff.ESPdata[zP];
  }
}

// --------------------------------------------------------------------------------
