// ################################################################################
//
//  MandalBot_BLE vR0
//
//  Released:  29/10/2023
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    Micro:            ESP32 Dev Module
    Partition scheme: Huge App (3MB)

    This code controls a self-balancing robot, which uses stepper motors as the drive
    mechanism, and is controlled using either Bluetooth BLE from a MagicseeR1 R1 Ring
    Controller, or over Wi-Fi using ESP-NOW, with a Wii Nunchuk (default). THese
    options are selected on power-up or reset.

    Base on the original NANO design, the PID runs at 4ms intervals (250 Hz), and the
    interrupt service routine (Isr) which generates the stepper driver clocks, has an
    interval of 40 microseconds (2.5 kHz).

    You can select different output drive calculations, to compare the performance of
    each approach. One being with a hard dead-band, around the balance point, and the
    other taking a more smoother appraoch.
*/
// Declare Libraries
#include <Arduino.h>          // needed when using an ESP32 micro
#include <Wire.h>             // include the Wire.h library so we can communicate with the gyro
#include <FastLED.h>          // Neopixel library
#include "BLEDevice.h"

#include <esp_now.h>          // ESP-NOW WiFi link library
#include <WiFi.h>             // ESP WiFi library
#include "SparkFun_VL53L1X.h" // VL53L1X laser ranger library

// Configuration
String Release = "MandalBot BLE (R0)";
String Date = "29/10/2023";

// Define a task to run on core 0
TaskHandle_t Core0;

//------ R1 Controller Definitions -----
enum {
  R1_TRIGGERS = 0,
  R1_JOYX,
  R1_JOYY,
  R1_BTNAB,
  R1_BTNCD,
  R1_NUMBYTES
};

// define constants for Display Monitor app
#define BLACK 0
#define BLUE 5
#define GREEN 4
#define INVERSE 2
#define MAGENTA 8
#define RED 3
#define TEXT_ALIGN_LEFT 0
#define TEXT_ALIGN_RIGHT 1
#define TEXT_ALIGN_CENTER 2
#define TEXT_ALIGN_CENTER_BOTH 3
#define TURQUOISE 7
#define WHITE 1
#define YELLOW 6

// BLE R1 Button Masks+
// Note that the Magicsee R1 needs to be placed in mouse mode by pressing M+D
// once it is powered up.
#define FRESHFLAG         0x80
#define JOYTIMEOUT        45      // joystick no activity timeout in mS, must be > 40ms for R1 controller
#define JoyStickDeadZone  0       // joystick centre position value
#define R1_LOW_TRIGGER    0x01    // R1 front C button, lower trigger
#define R1_UPR_TRIGGER    0x02    // R1 front Z button, upper trigger
#define R1_BUTTON_A       0x01    // R1 A button
#define R1_BUTTON_B       0x02    // R1 B button
#define R1_BUTTON_C       0x01    // R1 C button
#define R1_BUTTON_D       0x02    // R1 D button
#define ServerName  "Magicsee R1       "  // note 7 spaces after name must be included for BLE registration

// Define general constants
#define AcAvgDivMax 500           // max averaging divider, 500 == 5 second average
#define AccOffX 0                 // calibration offset applied to X-axis accelerometer (0)
#define AccOffY -264              // calibration offset applied to Y-axis accelerometer (0)
#define AccOffZ 1078              // calibration offset applied to Z-axis accelerometer (159)
#define AcRamp 30                 // rate limit for acceleration filtering
#define BatCal   287.7            // calibrated voltage multiplying factor
#define BatCritical 2895          // critical battery threshold 10.0v, 0%
#define BatHigh   3402            // high charge, voltage == 11.4v, 80%
#define BatMax    3841            // A0 for fully charged battery voltage, == 12.3v, 100%
#define BatPin      36            // analog input pin ADC15 on GPIO12
#define BatWarn   3175            // battery threshold of 10.5v gives low warning, 20%
#define BUILTINLED   2            // GPIO 2  -> built-in blue LED, on ->high, off -> low
#define I2Cdel 20                 // I2C bus timing delay for MPU6050
#define INT_PIN 0                 // VL53L1X GPIO interrupt to ADC input not actually used
#define IsrPrescale 40            // set the 80MHz prescaler to 40 giving StepTimer 2MHz clock
#define IsrTimeCnt  80            // set Timer limit, giving 40µs ISR interrupts
#define IsrTimerNum 1             // timer used in creating stepper ISR
#define LtofDet 350               // max range for target detection in head scanning
#define LtofGnd 350               // caliubration range for ground reflection
#define LtofLimit 500             // Range max value is always limited to this figure
#define LtofMax 400               // out of range value for head scanning
#define LtofMin 30                // minimum range value
#define LtofStop 140              // range value at which autonomous movement stops
#define EyeLED_Pin   5            // GPIO for RGB LED
#define GunLED_Pin  17            // GPIO for RGB LED
#define GyAvgDivMax 500           // max averaging divider, 500 == 5 second average
#define GyrOffX    234            // calibration offset taken from X-axis gyroscope (0)
#define GyrOffY    152            // calibration offset taken from Y-axis gyroscope (0)
#define GyrOffZ    230            // calibration offset taken from Z-axis gyroscope (0)
#define M0_Pin  27                // GPIO for DRV8825 M0 pin
#define M1_Pin  26                // GPIO for DRV8825 M1 pin
#define M2_Pin  25                // GPIO for DRV8825 M2 pin
#define MPU_address 0x68          // MPU-6050 I2C address (0x68 or 0x69)
#define NumLEDs      7            // max number of RGB LEDs used in one strip
#define TonePin      4            // PWM audio and tones will be sent to GPIO4
#define PwmChTone    7            // PWM channel assigned to playing tones
#define SafeCntMax 600            // count need to switch SafeMode from 0 to 1, in 4ms increments
#define SafeCntMax1 SafeCntMax/3  // count need to switch SafeMode from 0 to 1, in 4ms increments
#define SafeCntMax2 SafeCntMax1*2 // count need to switch SafeMode from 0 to 1, in 4ms increments
#define SCL         22            // GPIO 22 -> I2C SCL
#define SDA         21            // GPIO 21 -> I2C SDA
#define StepLftDir  19            // left stepper motor drive direction
#define StepLftEn   16            // left stepper motor driver enable, active LOW
#define StepLftStep 18            // left stepper motor driver step clock
#define StepRhtDir  32            // right stepper motor drive direction
#define StepRhtEn   13            // right stepper motor driver enable, active LOW
#define StepRhtStep 14            // right stepper motor driver step clock
#define sw0Pin      33            // GPIO 33 used for button switch SW0
#define sw1Pin       2            // GPIO 27 used for button switch SW1 - ####### this is temporary test
#define toppleAng 30.0            // angle to trigger safe mode topple
#define XSHUT_PIN   23            // GPIO 23 controls VL53L1X XSHUT HIGH enable pin

hw_timer_t *StepTimer = NULL;

// create a FastLED instance and define the total number of LEDs on the robot
// because the weapon and eye LEDs will be output at the same time we define the
// same number of LEDs, even though we only use 5 for the weapons and nose LEDs
// the gun LEDs are declared first as array [0] and eyes as array [1]
CRGB Guns_LED[NumLEDs];           // 5 weapon + nose LEDs connected to micro for o/p
CRGB Eyes_LED[NumLEDs];           // 7 eye LEDs connected to micro for o/p
CRGB Temp_LED[8];                 // temp store used in LED animation

// create VL53L1X range sensor instance, and define its pins
// Note that INT_PIN is declared for the library, but not used in the code
SFEVL53L1X VL53L1X(Wire, XSHUT_PIN, INT_PIN);

// ===== Magicsee R1 Modes =====
// This code assumes you are using the Mouse Mode, but by default R1 is in 
// @ + A -> Music & Video mode
// @ + B -> Horizontal Game mode
// @ + C -> Vertical Game mode
// @ + D -> Mouse Mode  // use this mode
//  4 byte notification, Trigger Sws, Joystick X, Joystick Y, 0x00

// All four modes send data. However each mode uses different byte positions and
// values for each of the switches. The joystick acts like a 'D' pad when not in
// Mouse Mode (no analog value).
 
typedef void (*NotifyCallback)(BLERemoteCharacteristic*, uint8_t*, size_t, bool);

// this is the service UUID of the R1 Control handheld mouse/joystick device (HID)
static BLEUUID serviceUUID("00001812-0000-1000-8000-00805f9b34fb");

// Battery Service UUID
static BLEUUID BatteryServiceUUID("0000180F-0000-1000-8000-00805f9b34fb");

// this characteristic UUID works for joystick & triggers (report)
static BLEUUID ReportCharUUID("00002A4D-0000-1000-8000-00805f9b34fb"); // report

// define BLE flags
static boolean doConnect = false;
static boolean connected = false;
static boolean doScan = false;
// static BLERemoteCharacteristic* pRemoteCharacteristic;
static BLEAdvertisedDevice* myDevice;
// static BLERemoteCharacteristic* pBatRemoteCharacteristic;
unsigned long BLEt0 = millis();
unsigned long BLEt1 = millis();
bool Button_A = false;
bool Button_B = false;
bool Button_C = false;
bool Button_D = false;

// pointer to a list of characteristics of the active service,
// sorted by characteristic UUID
std::map<std::string, BLERemoteCharacteristic*> *pmap;
std::map<std::string, BLERemoteCharacteristic*> :: iterator itr;

// pointer to a list of characteristics of the active service,
// sorted by characteristic handle
std::map<std::uint16_t, BLERemoteCharacteristic*> *pmapbh;
std::map<std::uint16_t, BLERemoteCharacteristic*> :: iterator itrbh;

// storage for pointers to characteristics we want to work with
// to do: change to linked list ?
BLERemoteCharacteristic *bleRcs[4];

// This is where we store the data from the buttons and joystick
volatile byte   R1_Data[R1_NUMBYTES];
volatile bool   flag = false;         // indicates new data to process
// joyTimer is a 30 millisecond re-triggerable timer that sets the joystick 
// back to center if no activity on the joystick or trigger buttons. 
volatile uint32_t joyTimer = millis();
char bfr[80];   // character buffer for UUID print statements


// Create ESP-NOW Wi-Fi structure
uint8_t broadcastAddress[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
// 250 byte max for Tx/Rx data
// It must match the receivers data structure
typedef struct ESPNOW_Buff250 {
    char ESPdata[250];
} ESPNOW_Buff250;
// Create an 8 byte message called NunchukDataTx, to hold I2C data messages
ESPNOW_Buff250 Tx_Buff;
// Create an 8 byte message to hold incoming sensor readings
ESPNOW_Buff250 Rx_Buff;

// PID controller factors, determined from PID tuning
// these are set later in code with a call to setPIDgains()
float Ptch_p_gain;      // Gain setting for Pitch P-controller (1.0)
float Ptch_i_gain;      // Gain setting for Pitch I-controller (0.0)
float Ptch_i_max;       // limit Pitch I-controller output(2.0)
float Ptch_d_gain;      // Gain setting for Pitch D-controller (0.0)
float Ptch_out_max;     // limit Pitch PID output to limit to PWM max
float PtchStartAng;     // initial target angle for balancing Pitch

// Declare and initialise global variables
long AcAvgDiv;                  // dividing factor used in averaging
int16_t AccDif;                 // difference between subsequent readings of AcY
bool AccOffEn;                  // if == false then display RAW Acc values
int16_t AccRampX;               // max allowable rate of change of AcX
int16_t AccRampY;               // max allowable rate of change of AcY
int16_t AccRampZ;               // max allowable rate of change of AcZ
int16_t A_Cnt;                  // counter used Test mode
int16_t AcX;                    // MPU raw X-acceleration value
long AcXAvg;                    // averaged AcX value to reduce noise
long AcXAvgL;                   // long averaged AcX value to reduce noise
int16_t AcXL;                   // previous MPU X-acceleration value
long AcXSum;                    // averaging accumulator
long AcXSumL;                   // long averaging accumulator
int16_t AcY;                    // MPU raw Y-acceleration value
long AcYAvg;                    // averaged AcY value to reduce noise
long AcYAvgL;                   // long averaged AcY value to reduce noise
int16_t AcYDwnCnt;              // counter used in orientation auto-mode switching
int16_t AcYL;                   // previous MPU Y-acceleration value
long AcYSum;                    // averaging accumulator
long AcYSumL;                   // long averaging accumulator
int16_t AcYUpCnt;               // counter used in orientation auto-mode switching
int16_t AcZ;                    // MPU raw Z-acceleration value
int16_t AcZd;                   // Z-acceleration difference between consequative readings
int16_t AcZi;                   // instantaneous Z-acceleration reading
long AcZAvg;                    // averaged AcZ value to reduce noise
long AcZAvgL;                   // longaveraged AcZ value to reduce noise
int16_t AcZL;                   // previous MPU Z-acceleration value
long AcZSum;                    // averaging accumulator
long AcZSumL;                   // long averaging accumulator
String Any$;                    // temporary string
int16_t AnyInt;                 // temporary variable
long AnyLong;                   // any temp long value
String Auton$;                  // autonomous mode string
int16_t AutoOFF;                // auto-power OFF timer
long AvCnt;                     // Av. filter width
int16_t BackAwayCnt;            // timer/counter used in 'Back Away' tasks
int16_t BackAwayTask;           // back away function task pointer
int16_t BackAwayTime;           // timer used in 'Back Away' function
int32_t BatAvg;                 // preload averaging function with high value
int16_t BatAvgMax;              // maximum average value in this period
byte BatCnt;                    // pointer to array values and critical timer
int16_t BatCritCnt;             // timer used in baterycritical testing
int16_t BatData[60];            // battery averages at 1 second intervals
int16_t BatDel;                 // set a 1 second delay in 'life' prediction
int16_t BatDP;                  // battery data pointer, 0 - 59
int16_t BatLast;                // previous battery average reading
int16_t BatPc;                  // battery percentage of charge
int16_t BatQ;                   // a measure of battery quality 1 - 5
int32_t BatSum;                 // preload averaging sum with 20x max value
unsigned long BatT0;            // time of first battery measurement in life prediction forecast
String BatTime$;                // time left displayed
unsigned long BatTms;           // time when battery reading was taken
int16_t BatV;                   // battery voltage * 10, as in 10.0v == 100
int32_t BatV0;                  // reference voltage used in prediction forecast
int32_t BatV1;                  // lowest voltage used in prediction forecast
float BatVfp;                   // battery voltage as a decimal number. ie. 11.2v
float BatVfpIni;                // battery voltage used to initialise SimpleFOC
int32_t BatVol;                 // instantaneous battery voltage
int16_t Beep;                   // makes a beep sound for a specified period
uint8_t Bright0;                // FastLED 0 brightness
uint8_t Bright1;                // FastLED 1 brightness
uint8_t Brightness;             // Display brightness
int16_t C_Cnt;                  // counter used in 'C' button detection
int16_t C_Dn;                   // counter for 'C' button down period
char cmdMode;                   // command mode
int16_t cmdSgn;                 // cmdVal sign, 0 or -1
char cmdType;                   // command mode type
int32_t cmdVal;                 // value associated with a cmdType
byte Col_B,Col_G,Col_R;         // current LED colours
byte C_Up;                      // counter for 'C' button up period
int16_t CZ;                     // received button values
bool CZ_Wait;                   // flag used for button release wait function
char Deg = 176;                 // degrees character in Monitor+
bool DEMO;                      // if == true then Demo Mode has been triggered
bool DispBack;                  // flag used to draw a back arrow on the display
unsigned long DispBatWarn;      // timer usesd to prevent battery warning repeating too often
bool DispChng;                  // flag invokes initialisation code
bool DispClr;                   // set == true to clear the memory contents
int16_t DispCnt;                // counter used in display updating
int16_t DispCX,DispCY;          // display cursor coordinates
int16_t DispDel;                // >0 display delay counter, stops updates over I2C
bool DispDisp;                  // set = true to immediately display memory contents
int16_t DispH;                  // display rectangle width
byte DispMenu;                  // group type, 0 = sensors, 1 = tasks, etc
int16_t DispMode;               // the type of display for a given mode
int16_t DispModeCnt;            // counter used in display updating
bool DispMon;                   // == true if connected to OLED Mirror app
int16_t DispNext;               // used to transition to another DispMode
bool DispNOP;                   // set == true to stop display I2C updates
int16_t DispNow;                // if > 0 display special modes
int16_t DispNxtCnt;             // counter used in auto-branching of display
unsigned long DispOverun;       // timer used to prevent the display function being over-run
int16_t DispPnt;                // points at the current display mode
int16_t DispTask;               // task pointer for display functions
int16_t DispTx;                 // counter used with display mirror app
int16_t DispVal;                // general display value, passed from tasks
int16_t DispW;                  // display rectangle width
int16_t DispX,DispY;            // co-ordinates for mirror app mouse click
int16_t DmB,DmX,DmY;            // button and X,Y values associated with a display monitor mouse click
int16_t DM_Control;             // 'Control' display assigned Display Mode
int16_t DM_DbBrake;             // 'DeadBand & Brake' display assigned Display Mode
int16_t DM_Drive;               // 'Drive' display assigned Display Mode
int16_t DM_Filter;              // 'LP Filter' display assigned Display Mode
int16_t DM_GunFire;             // 'Gun Fire' display assigned Display Mode
int16_t DM_Min,DM_Max;          // min/max Display Mode pointer values
int16_t DM_Motor_Stat;          // 'Motor Stats' display assigned Display Mode
int16_t DM_Motor_Test;          // 'Motor Test' display assigned Display Mode
int16_t DM_MPU_Acc;             // 'MPU Acc.' display assigned Display Mode
int16_t DM_MPU_AccRamps;        // 'Acc Ram,p.' display assigned Display Mode
int16_t DM_MPU_Gyros;           // 'MPU Gyros' display assigned Display Mode
int16_t DM_OpMode;              // 'Motor Calcs' display assigned Display Mode
int16_t DM_PtchPID;             // 'Pitch PID' display assigned Display Mode
int16_t DM_Range;               // 'Range' display assigned Display Mode
int16_t DM_SafeMode;            // 'Safe Mode' display assigned Display Mode
float Drive_Max;                // max value added to the PID_Op when driving
float Drive_Op;                 // value added to the PID_Op when driving
bool EnStepLft;                 // if == true then enable left stepper clock
bool EnStepRht;                 // if == true then enable right stepper clock
int16_t ESP_NOW_BA;             // broadcast address pointer, -1 to start with
bool ESP_NOW_Init;              // == true once ESP-NOW has been initiaiised
int16_t Gear;                   // speed control flag
uint8_t gBrightness = 255;      // brightness used in FastLED show events
int16_t GPmax,GPmin;            // start and end of graphing list, for Scope mode
int16_t GPnn;                   // pointer to current graphing items
bool GP_Title;                  // if == true then report a title in Scope graphing mode
int16_t GunFire;                // > 0 when firing the guns
long GyAvgDiv;                  // dividing factor used in averaging
int GyCal;                      // counter used to determine gyro calibration period
bool GyrOffEn;                  // if == false then display RAW Gyro values
unsigned long GyT;              // time of gyro readings in milliseconds
unsigned long GyTD;             // time between successive gyro readings in milliseconds
int16_t GyX;                    // MPU raw X-gyroscope value
long GyXAvg;                    // averaged GyX value to reduce noise
long GyXAvgL;                   // long averaged GyX value to reduce noise
int16_t GyXL;                   // previous MPU X-gyroscope value
int16_t GyXOff;                 // offset to be added to raw X-gyroscope value
int16_t GyXOffCnt;              // offset counter for gyro centering
long GyXSum;                    // averaging accumulator
long GyXSumL;                   // long averaging accumulator
int16_t GyY;                    // MPU raw Y-gyroscope value
long GyYAvg;                    // averaged GyY value to reduce noise
long GyYAvgL;                   // long averaged GyY value to reduce noise
int16_t GyYL;                   // previous MPU Y-gyroscope value
int16_t GyYOff;                 // offset to be added to raw Y-gyroscope value
int16_t GyYOffCnt;              // offset counter for gyro centering
long GyYSum;                    // averaging accumulator
long GyYSumL;                   // long averaging accumulator
int16_t GyZ;                    // MPU raw Z-gyroscope value
long GyZAvg;                    // averaged GyZ value to reduce noise
long GyZAvgL;                   // long averaged GyZ value to reduce noise
int16_t GyZL;                   // previous MPU Z-gyroscope value
int16_t GyZOff;                 // offset to be added to raw Z-gyroscope value
int16_t GyZOffCnt;              // offset counter for gyro centering
long GyZSum;                    // averaging accumulator
long GyZSumL;                   // long averaging accumulator
byte I2CErr;                    // flag returned from I2C comms
bool I2C_LTOF;                  // true if VL53L1X is detected during POST
bool I2C_MPU;                   // true if MPU6050 is detected during POST
volatile bool IsrRun = false;   // if == true then run ISR code, default = false
bool IsrRunF;                   // a temporary copy of IsrRun, when IsrRun is switched OFF by a function
volatile int IsrT0,IsrT1;       // ISR performance timers used in development
int IsrTimer = 0;               // one of four hardware timer used for interrupts 0-3
int16_t JoyX;                   // received JX value
int16_t JoyY;                   // received JY value
char keyChar;                   // any keyboard character
int16_t keyVal;                 // any keyboard value
unsigned long LED20ms;          // used to generate 20ms LED tasks in Core 1
bool LED40ms;                   // a 40ms timer derived from the LED20ms event
int16_t LED40msTask;            // 40ms tasks running on Core 0
int16_t LED0_Cnt;               // general counter used in LED tasks
int16_t LED1_Cnt;               // general counter used in LED tasks
int16_t LED0_Del;               // gun LED Task delay, adds pauses into the processes
int16_t LED1_Del;               // eye LED Task delay, adds pauses into the processes
int16_t LED0_Flash;             // if > 0 then flash Gun LEDs bright white for n counts
int16_t LED1_Flash;             // if > 0 then flash Eye LEDs bright white for n counts
int16_t LED0_Last;              // record previous LED0 task
int16_t LED1_Last;              // record previous LED1 task
int16_t LED0_New;               // pointer to new Gun LED task
int16_t LED1_New;               // pointer to new Eye LED task
bool LED0_Set;                  // flag used to initiate Gun LED task change
bool LED1_Set;                  // flag used to initiate Eye LED task change
int16_t LED0_Task;              // Gun LED task pointer
int16_t LED1_Task;              // Eye LED task pointer
bool LED0_Flg;                  // LED flag used in simple flashing mode
bool LED1_Flg;                  // LED flag used in simple flashing mode
int16_t LED0_Inc;               // LED Task increment variable
int16_t LED1_Inc;               // LED Task increment variable
int16_t LED0_Pnt;               // LED pointer variable
int16_t LED1_Pnt;               // LED pointer variable
bool LED0_Run;                  // == true if face LED_Tasks are to run
bool LED1_Run;                  // == true if gun LED_Tasks are to run
bool LEDshow0 = false;          // if true then update Guns_LED strip
bool LEDshow1 = false;          // if true then update Eyes_LED strip
int16_t LED0_SubTsk;            // LED sub-task pointer
int16_t LED1_SubTsk;            // LED sub-task pointer
int16_t LED_Tasks;              // LED 20ms task pointer
bool Lft_Dir;                   // left motor DIR pin status
bool Lft_En;                    // left motor enable, active LOW, default = HIGH
int Lft_Motor;                  // left motor demand, adjusted for throttle
int Lft_Motor_Cnt;              // left motor counter
int Lft_Motor_Tgt;              // left motor target count
int Lft_Motor_Tgt2;             // left motor target count/2
volatile int Lft_Motor_Throttle;  // left motor throttle demand
float LftPtch_Op;               // PID output for left motor
int Lft_Step_Cnt;               // left motor step counter
int Lft_Step_Inc;               // left motor step increment
float LpFc;                     // low pass filter co-efficient, between 1.00 to 0.01 (min/max)
bool LTOFcalc;                  // if == true then calculate speed of ranged object in m/s
int16_t LTOFfps;                // ranger sensor measurement rate in fps
int16_t LTOFperiod;             // range sensor inter-measurementg period in millis()
int16_t LTOFspd10;              // speed of ranged target in m/s x 10
int16_t LTOFspd10L;             // previous speed of ranged target in m/s x 10
int16_t LTOFspdCnt;             // speed peak detector sustain limit
int16_t LTOFt0,LTOFt1;          // range sensor timers for fps
uint8_t M210;                   // driver m pin assignments
int16_t MainMode;               // determined after RESET, based on orientation
int16_t MainType;               // a number representing the type of task, default = 0
bool Motor_Test;                //  if == true then conduct motor tests, default = false
int16_t moveDir;                // move direction flag. 0 == stationary
String myMAC;                   // MAC address if this device
String MyStatus;                // status displayed in MOnitor+ app
unsigned long next4ms;          //  4ms (250Hz) primary loop timer
unsigned long nextLftStep;      // left motor step timing trigger
unsigned long nextRhtStep;      // right motor step timing trigger
note_t NoteLast;                // most recently played note
uint8_t Octave;                 // most recently played octave
int16_t OpMode;                 // determines the output calculation method, 0, 1,etc
int16_t Ping;                   // Rx 'ping' counter
int16_t PrintDel;               // delay counter used in serial printing
float PtchAccAng;               // Pitch angle measured by the Y accelerometer
float Ptch_Brake;               // brake factor applied to PID output
float Ptch_D;                   // Pitch PID D-gain component
float Ptch_Db;                  // Pitch deadband limit in terms of Pitch_Op
bool Ptch_d_En;                 // Pitch PID d-gain enable true=ON, false=OFF
bool Ptchen;                    // if == true then Pitch PID mode is enabled
float PtchErr;                  // current Pitch PID error
bool PtchGo;                    // == true when Pitch axis is within starting angle
float PtchGyrAng;               // angle measured by Y gyroscope and corected, as true angle
float PtchGyrAngLast;           // previous angle measured by Y gyroscope and corrected, as true angle
float PtchLastErr;              // previous Pitch PID error, used in D calculations
bool Ptch_i_En;                 // Pitch PID i-gain enable true=ON, false=OFF
float Ptch_i_mem;               // used in Pitch PID integral calcs
float PtchLastDErr;             // used in Pitch PID derivative calcs
int PtchNan;                    // PItch 'Not A Number' error count
float Ptch_Op;                  // Pitch PID output
float Ptch_OpBrake;             // PID output at which the brake function is applied
float Ptch_P;                   // Pitch PID p-gain component
float PtchSp;                   // primary Pitch PID setpoint
float PtchSpTgt;                // target setpoint for controller demands, Gear dependant
int16_t PtchStepT;              // Pitch impulse step test duration in x5ms
unsigned long print40ms;        // 40ms task pointer - variable
int16_t PrintTgt;               // 0 == serial port, 1 == WiFi ESP-NOW
String PrintTx = "";            // printed strings
float Pwr;                      // PID output power setting, 0.5 to 1.0
int Range;                      // range value limited to RangeMax
bool RangeEn;                   // == true if VL53L1X ranging is active
int16_t RangeCnt;               // discriminator counter
int16_t RangeCntMax;            // discriminator counter, adjustable limit
int RangeDiff;                  // difference in RangeRaw from RangeLast
bool RangeFps;                  // == true if LTOF fps is to be displayed
int RangeLast;                  // previous range measurement
bool RangeLL;                   // == true if range < RangeMin
int RangeMax;                   // the current limited to Range measurements
int RangeMin;                   // the current lower limit of Range measurements
bool RangeNew;                  // set == true when a new range is available
int RangePkPk;                  // Range pk-pk noise limit (default 20)
int RangeRate;                  // max rate of inter-measurement Range change, default = 64
int RangeRaw;                   // range value obtained from VL53L1X
int RangeRawF;                  // rate limited RangeRaw value
int RangeTrkHi;                 // range tracking high value
int RangeTrkLo;                 // range tracking low value
bool RangeUL;                   // if == true if range exceeds RangeMax
bool Ready;                     // if == true then we have completed setup tasks, default = false
bool readWiiCall;               // flag prevents call stack overflow
long Report20ms;                // timer used in serial port reporting, for debug
bool Rht_Dir;                   // right motor DIR pin status
bool Rht_En;                    // right motor enable, active LOW, default = HIGH
int Rht_Motor;                  // right motor demand, adjusted for throttle
int Rht_Motor_Cnt;              // right motor counter
int Rht_Motor_Tgt;              // right motor target count
int Rht_Motor_Tgt2;             // right motor target count/2
volatile int Rht_Motor_Throttle;  // right motor throttle demand
float RhtPtch_Op;               // PID output for right motor
int Rht_Step_Cnt;               // right motor step counter
int Rht_Step_Inc;               // right motor step increment
float RollAccAng;               // Roll angle measured by the Z accelerometer
float RollGyrAng;               // angle measured by Z gyroscope and corected, as true angle
uint8_t RxByte;                 // byte containing most recent WiFi control data
int16_t Rx_Chk;                 // Rx checksum
bool RxClass;                   // set == true when a Wii Classic command has been receieved
int16_t RxCZ;                   // buffered received CZ value
int16_t RxJoyX;                 // buffered received JoyX value
int16_t RxJoyY;                 // buffered received JoyY value
int16_t Rx_len;                 // length of WiFi received data block
bool RxNunch;                   // set == true when a Wii Nunchuk command has been receieved
int16_t Rx_Pnt;                 // data pointer user by Rx state machine
bool RxRec;                     // set true when a valid frame of data is received
int16_t RxState;                // receiver state machine state
int16_t Rx_Task;                // task pointer user by Rx state machine
int16_t RTxTimeout;             // counter used to auto-reset if WiFi fails
int16_t RxVal;                  // value received from serial Rx
byte RxWiFi[6];                 // array holding 6 bytes of Wii received data
int16_t SafeCnt;                // counter used in Safe Mode checks
int16_t SafeMode;               // controls the state of the control system
bool Sb;                        // if == true then perform self-balance setpoint adjustments (true)
// int16_t SbCnt;                  // counter used in Pitch auto self-balance point adjustment
// int16_t Sb_trend;               // counter used in Pitch auto self-balance point adjustment
float SbSp;                     // Pitch PID self-balance set points
float SbSpInc;                  // Pitch self-balance auto-increment
int16_t ScpDpth;                // last item reference in the scope table
int32_t ScpPnt0 = 0;            // table pointer for scope trace 0
int32_t ScpPnt1 = 1;            // table pointer for scope trace 1
int32_t ScpPnt2 = 2;            // table pointer for scope trace 2
int32_t ScpPnt3 = 3;            // table pointer for scope trace 3
int32_t ScpPP = 0;              // pointer to the target ScpPnt0-3 variable
String ScpTitle$;               // selected title in the scope table
String ScpVal$;                 // selected value in the scope table
bool SerialRx;                  // == true if any character has been received
String Status$ = "";            // any status message to be displayed on OLED
bool StepLft;                   // left stepper driver step clock
bool StepLftHi;                 // == true if left stepper driver step clock is HIGH
unsigned long StepLftPeriod;    // left stepper clock period in µs
bool StepRht;                   // right stepper driver step clock
bool StepRhtHi;                 // == true if right stepper driver step clock is HIGH
unsigned long StepRhtPeriod;    // right stepper clock period in µs
int16_t sw0Cnt;                 // button switch counter
int16_t sw0DwnTime;             // button pressed down time
bool sw0LastState;              // previous state of button switch, HIGH/LOW
bool sw0_Nop;                   // if == true don't perform switch functions
bool sw0State;                  // state of read button switch pin
int16_t sw0Timer;               // timer used to detemine button sequences
bool sw0Wup;                    // set == true for SW0 to wait for button release
int16_t sw1Cnt;                 // button switch counter
int16_t sw1DwnTime;             // button pressed down time
bool sw1LastState;              // previous state of button switch, HIGH/LOW
bool sw1_Nop;                   // if == true don't perform switch functions
bool sw1State;                  // state of read button switch pin
int16_t sw1Timer;               // timer used to detemine button sequences
bool sw1Wup;                    // set == true for SW1 to wait for button release
unsigned long t0,t1;            // debug timers
unsigned long t2,t3;            // debug timers
int16_t Task4ms;                // 4ms task pointer
int16_t Task8ms;                // 8ms task pointer
int16_t Task20ms;               // 20ms task pointer
int16_t Task40ms;               // 40ms task pointer
bool TEST = true;               // set == true during assembly, then false for normal use
int16_t TestMode;               // controls the state of the control system
bool ToneAtt;                   // if == true then tone channel is attached, default = false
int16_t ToneDel;                // tone delay counter
int ToneFreq;                   // frequency used in TonyPlay() function
int16_t TonePnt;                // tone task pointer
int16_t Turning;                // value used to track applied controller demands
float TurnSpd;                  // value applied to right/left when turning
float TurnSpdInc;               // increment applied to right/left turning rate
float TurnSpdMax;               // max value applied to right/left when turning for that Gear
float TurnSpdMin;               // max value applied to right/left when turning for that Gear
int16_t Tx_PingCnt;             // 1 second Tx ping timer
bool USB;                       // == true if initial battery is below 5.5v
uint8_t VL53L1X_OC;             // VL53L1X ROI SPAD centre
uint16_t VL53L1X_ROIX;          // VL53L1X ROI SPAD width
uint16_t VL53L1X_ROIY;          // VL53L1X ROI SPAD height
unsigned long VL53L1X_Skip;     // read function skip timer
unsigned long VL53L1X_SkipDel;  // read function skip timer delay tracker
int16_t VL53L1X_Status;         // tracks value of status register
int16_t VL53L1X_Task;           // if > 0 then run VL53L1X tasks
unsigned long VL53L1X_Timer;    // general timer
int16_t WiFiCntC;               // C button counter for WiFi enable
int16_t WiFiConCnt;             // counter used in connection retry times
bool WiFiConnected;             // == true if a WiFi connection has been established
bool WiFiEn;                    // == true to enable WiFi control
bool WiFiMode;                  // if true then ESP-NOW, else BLE
int16_t WiFiPing;               // Rx counter used to maintain connected state
int16_t WiFiRestCnt;            // counter used to go to rest if no WiFi demands
bool WiFiRx;                    // true when a block of data has been received
long WiFiRxBytes;               // number of bytes received over the WiFi link
int16_t WiFiRxCnt;              // counter used to reset controller demands when disconnected
int16_t WiFiRxErr;              // number of bytes received over the WiFi link
bool WiFiRxRec;                 // true when a block of data has been successfully received
int16_t WiFiTryCnt;             // WiFi try to connect count on error count
int16_t WiFiTryNum;             // WiFi try to connect total count
bool WiFiTx;                    // true when a block of data has been sent successfully
int16_t WiFiTx_CB;              // == 1 call back is clear to send, == -1 resend
int16_t WiFiTx_Chk;             // Tx checksum
int16_t WiFiTx_len;             // >0 when WiFi Tx buffer contains data to be sent
long WiFiTxBytes;               // number of bytes received over the WiFi link
int16_t WiFiTxErr;              // WiFi data error
int16_t WiFiTxErrCnt;           // WiFi Tx error count, leads to disconnection
uint8_t WiiBYAX;                // Wii Classic right-hand XYAB buttons
uint8_t WiiDigi;                // Wii Classic digital pad bits
uint8_t WiiSHS;                 // Wii Classic centre select,home,start bits
int16_t WiiType;                // device type, 0 == Nunchuk, 1 == Classic
float YawAccAng;                // Yaw angle measured by the X accelerometer
float YawGyrAng;                // angle measured by X gyroscope and corected, as true angle
float YawGyrAngLast;            // previous YawGyrAng
int16_t YawTimeout;             // timeout used in Yaw angle display
int16_t Z_Dn;                   // Z-button down count, 0 == released, default

// ---------------------------------------------------------------------------------

void IRAM_ATTR onTimer() {
  // this is the IsrTimer ISR routine, which is called every 40µs (25 kHz)
  // this code must be placed before the setup code for the compiler to recognise it
  // for the code to run IsrRun == true and variables must have been set beforehand
  // the code must be very efficient, as it is an overhead on everything else
  if (IsrRun) {
    // Left-hand motor, Lft_En == LOW to run stepping process
    if (!Lft_En && (Lft_Motor_Throttle != 0)) {
      Lft_Motor_Tgt = abs(Lft_Motor_Throttle);                // update the left target count
      Lft_Motor_Cnt ++;                                       // increase the left motor counter
      if (Lft_Motor_Cnt > Lft_Motor_Tgt) {                    // test for reaching target count
        Lft_Motor_Cnt = 0;                                    // reset the left motor counter
      }
      else if (Lft_Motor_Cnt == 1) {digitalWrite(StepLftStep,HIGH);}  // output rising edge of step pulse
      else if (Lft_Motor_Cnt == 3) {
        digitalWrite(StepLftStep,LOW);                        // output falling edge of step pulse
        Lft_Step_Cnt += Lft_Step_Inc;                         // increment the step counter
      }
    }

    // Right-hand motor, Rht_En == LOW to run stepping process
    if (!Rht_En && (Rht_Motor_Throttle != 0)) {
      Rht_Motor_Tgt = abs(Rht_Motor_Throttle);                // update the right target count
      Rht_Motor_Cnt ++;                                       // increase the right motor counter
      if (Rht_Motor_Cnt > Rht_Motor_Tgt) {                    // test for reaching target count
        Rht_Motor_Cnt = 0;                                    // reset the right motor counter
      }
      else if (Rht_Motor_Cnt == 1) {digitalWrite(StepRhtStep,HIGH);}  // output rising edge of step pulse
      else if (Rht_Motor_Cnt == 3) {
        digitalWrite(StepRhtStep,LOW);                        // output falling edge of step pulse
        Rht_Step_Cnt += Rht_Step_Inc;                         // increment the step counter
      }
    }
  }
}

// --------------------------------------------------------------------------------

class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
  //******************************************************************************
  // Scan for BLE servers and find the first one that advertises the service we are looking for.
  //******************************************************************************
  // Called for each advertising BLE server.
  // This is only used when in BLE mode, entered by pressing SW0 during boot.
  void onResult(BLEAdvertisedDevice advertisedDevice) {
    Serial.print("BLE Advertised Device found: ");
    Serial.println(advertisedDevice.toString().c_str());

    // we have found a server, see if it has the name we are looking for
    if (advertisedDevice.haveName()) {
      if (0 == strcmp(ServerName, advertisedDevice.getName().c_str())) {
        Serial.println("Found Magicsee R1 Server");

        // we found a server with the correct name, see if it has the service we are
        // interested in (HID)
  
        if (advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(serviceUUID)) {
          Serial.println("Server has HID service");
          BLEDevice::getScan()->stop();
    
          myDevice = new BLEAdvertisedDevice(advertisedDevice);
          doConnect = true;
          doScan = true;
        } else {
          // did not find the Magicsee HID server
          Serial.println("Server does not have an HID service, not our server");
        }
      }
    } else {
      Serial.println("Server name does not match, not our server");
    }
  } // onResult
};

// --------------------------------------------------------------------------------

class MyClientCallback : public BLEClientCallbacks {
  //******************************************************************************
  // Connection state change event callback handler.
  //******************************************************************************
  void onConnect(BLEClient* pclient) {
    Serial.println("onConnect event.");
  }

  void onDisconnect(BLEClient* pclient) {
    Serial.println("onDisconnect event.");
    connected = false;
  }
};

// --------------------------------------------------------------------------------

void setup() {
  // initialise pins and tasks
  // the first thing to do is turn off the stepper drivers
  pinMode(StepLftEn,OUTPUT); digitalWrite(StepLftEn,HIGH);      // left stepper enable pin, active LOW
  pinMode(StepRhtEn,OUTPUT); digitalWrite(StepRhtEn,HIGH);      // right stepper enable pin, active LOW

  pinMode(BatPin,INPUT);                                        // ADC1_0 GPIO36 battery divider
  pinMode(M0_Pin,OUTPUT); digitalWrite(M0_Pin,LOW);             // DRV8825 M0 pin
  pinMode(M1_Pin,OUTPUT); digitalWrite(M1_Pin,HIGH);            // DRV8825 M1 pin
  pinMode(M2_Pin,OUTPUT); digitalWrite(M2_Pin,LOW);             // DRV8825 M2 pin
  pinMode(StepLftDir,OUTPUT); digitalWrite(StepLftDir,LOW);     // left stepper direction pin
  pinMode(StepLftStep,OUTPUT); digitalWrite(StepLftStep,LOW);   // left stepper step pin
  pinMode(StepRhtDir,OUTPUT); digitalWrite(StepRhtDir,LOW);     // right stepper direction pin
  pinMode(StepRhtStep,OUTPUT); digitalWrite(StepRhtStep,LOW);   // right stepper step pin
  pinMode(sw0Pin,INPUT_PULLUP);                                 // button switch SW0 pin
  pinMode(sw1Pin,INPUT_PULLUP);                                 // button switch SW1 pin
  pinMode(XSHUT_PIN,OUTPUT); digitalWrite(XSHUT_PIN,LOW);       // switch VL53L1X OFF

  // take initial battery reading, to set ESP32 mux
  BatVol = analogRead(BatPin);
  
  resetDMs();                                 // reset display mode references
  setPIDgains();                              // set the default PID gains
  setDefaults();                              // load default values
  Serial.begin(115200); Serial.print("\n\n");
  Wire.begin();                               // start the I2C bus as master

  // initialise the FastLED component
  // a chain of 7 LEDs = 24 x 7 = 168 bits, taking 202us to show at 800kbps
  FastLED.addLeds<WS2812B, GunLED_Pin, GRB>(Guns_LED, NumLEDs);
  FastLED.addLeds<WS2812B, EyeLED_Pin, GRB>(Eyes_LED, NumLEDs);
  FastLED.setBrightness(Brightness);
  fill_solid(Eyes_LED, 7, CRGB::Black);
  fill_solid(Guns_LED, 7, CRGB::Black);

  // initialise I/O
  runPOST();

  if (!WiFiMode) {
    // BLE mode selected, so initialised interface
    Serial.println("BLE WiFi mode selected.");
    //#######################################################
    // Set up BLE
    //#######################################################
    Serial.println("Starting ESP32 BLE Client...");
    BLEDevice::init("");

    // Retrieve a GATT Scanner and set the callback we want to use to be informed 
    // when we have detected a new device.  Specify that we want active scanning
    // and start the scan to run for 5 seconds.
    // Set RGB LEDs to blue indicate scanning
    // LED[0].setRGB(0,0,20); LED[1].setRGB(0,0,20);  LED[2].setRGB(0,0,20); FastLED.show();
    // Serial.println(F("RGB LEDs are blue."));
    BLEScan* pBLEScan = BLEDevice::getScan();
    pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
    pBLEScan->setInterval(1349);
    pBLEScan->setWindow(449);
    pBLEScan->setActiveScan(true);
    pBLEScan->start(5, false);      // scan for 5 seconds

    if (!connected) {
      // if disconnected then restart scanning in loop()
      doScan = true;
      Serial.println("Offline, start a scan");
    }
  } else {
    // ESP-NOW WiFi mode selected, so initialised interface
    Serial.println("ESP-NOW WiFi mode selected.");
    //#######################################################
    // Set up ESP-NOW Wi-Fi
    //#######################################################
    // initialise ESP-NOW WiFi link
    Serial.println(F("RGB LEDs are green."));
    WiFi.mode(WIFI_STA);
    myMAC = WiFi.macAddress();
    Serial.println("my MAC= " + String(myMAC));
    WiFi.disconnect();
    // initialise the ESP-NOW link for the first time
    Init_ESP_NOW();
  }

  // define the Core 0 task parameters
  // Core 0 is normally exclusively used for WiFi, but here we use it to control
  // RGB LED patterns and clock them out
  // Note, this code starts and runs immediately, before loop() does
  xTaskCreatePinnedToCore(
    loop0,       // Task function.
    "Core0",     // name of task.
    10000,       // Stack size of task
    NULL,        // parameter of the task
    1,           // priority of the task
    &Core0,      // Task handle to keep track of created task
    0);          // pin task to core 0

  // Off we go
  synchLoopTimers();
  Isr_En();
}

// --------------------------------------------------------------------------------

void loop() {
  //###############################################################################
  //
  // Core 1 main loop functions
  //
  //###############################################################################
  // Check primary 4ms PID timer, and trigger all sub-tasks from that.
  // This is a multi-tasking environement, based on a 4ms timer, with subtasks
  // scheduled to run at that rate, or multiples of it, as in 8ms, 20ms, etc
  if ((millis() - next4ms) >= 4) {
    next4ms = millis(); Task4ms = 2;
    // run PID function as first task if enabled by SafeMode state
    if (SafeMode >= 3) {Pitch_PID();}
  }

  readSerial();                 // check serial port on every loop cycle

  if (Task4ms > 0) {
    // the 4ms timer has released the 4ms tasks to run
    switch (Task4ms) {
      case 1: 
        // every 4ms we update the sub task scheduler
        Task8ms--;  if (Task8ms  <= - 2) {Task8ms =  5;}
        Task20ms--; if (Task20ms <= - 5) {Task20ms = 4;}
        Task40ms--; if (Task40ms <= -10) {Task40ms = 5;}
        break;
      case 2: Safe_Mode(); break; // Safe Mode checks are performed every 4ms
    }
    Task4ms--;
  }

  if (Task8ms > 0) {
    // the are 8ms tasks to do
    //#############################################################################
    // 8ms tasks
    //#############################################################################
    // the following tasks are run in sequence, as a cascade.
    if (SafeMode >= 3) {Task8ms = 0;}                 // the PID function is running
    else {
      // Read MPU6050 when the PID controller is not active in self-balance mode
      switch(Task8ms) {
        case 1: if (I2C_MPU) {CalcAng();} break;        // vertical angle calc
        case 2: if (I2C_MPU) {AvgGyrAll();} break;      // average sensor Gyro values
        case 3: if (I2C_MPU) {AvgAccelAll();} break;    // average sensor Accel values
        case 4: if (I2C_MPU) {PID_readAcc();} break;    // read MPU6050 accelerometers
        case 5: if (I2C_MPU) {PID_readGyr();} break;    // read MPU6050 gyroscopes
      }
    } Task8ms--;
  }

  //#############################################################################
  // 20ms tasks
  //#############################################################################
  // These 20ms tasks run as a cascaded sequence once triggered every 20ms
  if (Task20ms > 0) {
    // tasks are performed in reverse order as the pointer is counting down
    switch(Task20ms) {
      case 1: readSW1(); break;
      case 2: readSW0(); break;
      case 3: if (SafeMode == 4) {BackAway();} break;
      case 4: // if in ESP-NOW WiFi mode, read Wii or auto cancel controller demand
         if (WiFiMode) {
           if (WiFiConnected) {readWii();}
           if (WiFiRxCnt > 0) {
             WiFiRxCnt --;
             if ((WiFiRxCnt < 1) && (BackAwayTask == 0)) {RxByte = 0b00000000;}}
         } break;
    }
    Task20ms--;
  }

  //#############################################################################
  // 40ms tasks
  //#############################################################################
  // these 40ms tasks run after the 20ms tasks are comnpleted
  if (Task40ms > 0) {
    // tasks are performed in reverse order as the pointer is counting down
    switch(Task40ms) {
      case 1: // if DispMon is active we run a time-out on the link
        if (DispTx > 0) {DispTx--; if (DispTx == 0) {DispMon = false;}}
        break;
      case 2: battRead(); break;
      case 3: if (Ping > 0) {Ping--;} break;      // time-out app data received flag
      case 4: // if not connected, in ESP-NOW WiFi, then try to connect every second
        if (WiFiMode) {
          // we are in ESP-NOW WiFi, not BLE mode
          if (!WiFiConnected) {
            // only try to connect WiFi if no 'Pings' are being received
            if (Ping == 0) {
              WiFiConCnt--;
              if (WiFiConCnt < 1) {
                WiFiConCnt = 50; WiFiTryToConnect();
              //  Serial.println("Trying to connect WiFi");
              }
            }
          }
        } break;
      case 5: // auto disconnect for normal WiFi mode
        if (WiFiMode) {
          // we are in ESP-NOW WiFi, not BLE mode
          if (WiFiPing > 0) {
            WiFiPing--;
            if (WiFiPing < 1) {
              // we have stopped receiving pings from WiFi controller,
              // so disconnect WiFi
              WiFiDisconnect();
              // Serial.println("WiFiPing == 0");
            }
          }
        } break;
    }
    Task40ms--; // decrement task pointer towards 0
  }

  //###############################################################################
  //  40 ms Print Task - Variable
  //###############################################################################
  // This code uses either the serial port or the WiFi link to send messages
  // at 115200 baud we can send 11.5 chars/ms, so 64 chars takes 5.5ms.
  if ((millis() - print40ms)  >= 40) {
    print40ms = millis();
    if (PrintTx.length() > 0) {
      // characters in string buffer so send some/all of them
      print40ms = millis() - 34;  // shorten the delay to 6ms for more data to send
      Tx_PingCnt = 0;             // reset the WiFi ping counter
      if ((!WiFiMode) || (PrintTgt == 0)) {
        // default uses the serial port
        if (Serial.availableForWrite() >= 64) {
          if (PrintTx.length() <= 64) {
            Serial.print(PrintTx);
            PrintTx = "";   // empty the buffer
          } else {
            Serial.print(PrintTx.substring(0,64));
            PrintTx = PrintTx.substring(64);
          }
        }
      } else {
        // text data has been received from WiFi so respond using ESP_NOW
        // ensure that we have had a call back from the previous frame
        if (WiFiConnected) {
          // only trannsmit PrintTx data if connected
          if (WiFiTx_CB == 1) {
            WiFiTx_CB = 0;  // clear the call back flag when sending
            if (PrintTx.length() <= 32) {
              WiFiTx_len = PrintTx.length();
              PrintTx.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
             t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = "";   // empty the buffer
            } else {
              WiFiTx_len = 32;
              Any$ = PrintTx.substring(0,WiFiTx_len);
              Any$.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
             t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = PrintTx.substring(32);
            }
          } else if (WiFiTx_CB == -1) {
            // an error has occured so resend the data
            if (WiFiTryCnt < 5) {
              WiFiTryCnt++;   // track the number of retries
              WiFiTx_CB = 0;  // clear the call back flag when sending
             t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
            } else {
              // too many resends so disconnect WiFi link
              WiFiDisconnect();
              Serial.println("Failed to Tx!");
            }
          }
        } else {
          // not connected so clear PrintTxdata
          PrintTx = "";
        }
      }
    }
  }

  if (!WiFiMode) {
    // WiFiMode determines whether we use an ESP-NOW or Bluetooth comms link.
    // The Monitor+ only works with the ESP-NOW system and Nunchuk controller.
    // Having both systems included uses up a lot of code space.
    //#######################################################
    // BLE checks
    //#######################################################
    // If the flag "doConnect" is true then we have scanned for and found the desired
    // BLE Server with which we wish to connect.  Now we connect to it.  Once we are 
    // connected we set the connected flag to be true.
    if (doConnect == true) {
      // report these messages once
      if (connectToServer()) {
        // Serial.println("We are now connected to the BLE Server.");
      } else {
        // Serial.println("We have failed to connect to the server; there is nothing more we will do.");
      } doConnect = false;
    }

    if (connected) {
      // joystick no activity detector
      if (joyTimer && (joyTimer < millis())) {
      //  Serial.println("Timeout");
        // no joystick notification for 45mS, centre the joystick
        R1_Data[R1_JOYX] = R1_Data[R1_JOYY] = 0;
        // call the joystick function
        HandleJS();
        joyTimer = 0;
      }
    } else if (doScan) {
      doScan = false;
      Serial.println("Start scanning after a disconnect, in loop().");
      // this is just example to start scan after disconnect, most likely there is
      // a better way to do this, as this approach halts the main loop() tasks.
      // so the robot will not respond until the Magicsee R1 controller is switched on
      BLEDevice::getScan()->start(0);
    }
  } else {
    //###############################################################################
    //  WiFi comms check & respond
    //###############################################################################
    // if WiFi is conencted then check Rx buffer on every cycle, so that we respond to
    // received data packets as soon as possible
    if (WiFiConnected) {
      if (WiFiRx) {
        // WiFi Rx buffer contains unread data so deal with it before adding new
        readWiFiRx();
      }
    } 
    // if not in Back Away mode then clear the control byte
    else if (BackAwayTask == 0) {RxByte = 0b00000000;}
  }

  // LTOF measurements are controlled and timed by the function call
  if (VL53L1X_Task > 0) {VL53L1X_Run();}  // perform laser measurement tasks
}

// ---------------------------------------------------------------------------------

void loop0(void * pvParameters) {
  // This special function runs on Core 0, also used for ESP32 WiFi NOW.
  // *** WARNING *** do not use for time consuming code!
  // The following message, sent once at boot, confirms core 0 assignment
  Serial.print("Core0 task running on core "); Serial.println(xPortGetCoreID());

  for(;;){
    // An infinite loop, just like loop() but running in Core 0.
    // You must never use 'break' or 'return' statement here or it will crash.
    // We tell the free RTOS system to call this function every 1 ms, but your tasks
    // need to be quite quick. Too much to do will crash the RTOS, so we do things in
    // cascade way.

    //###############################################################################
    //  Update RGB LEDs if needed
    //###############################################################################
    // Show events occur at 20ms intervals.
    // 1st priority is given to showing LED changes.
    if ((millis() - LED20ms) >= 20) {
      LED20ms = millis();
      // Check for a show event to synchronise LED updates
      if (LEDshow0 || LEDshow1) {
        if (GunFire) {
          FastLED[0].showLeds(Bright0); FastLED[1].showLeds(Bright0);
          LEDshow0 = false; LEDshow1 = false;
        } else if (LEDshow0 && LEDshow1) {
          FastLED[0].showLeds(Bright0); FastLED[1].showLeds(Bright1);
          LEDshow0 = false; LEDshow1 = false;
        } else if (LEDshow0) {
          FastLED[0].showLeds(Bright0); FastLED[0].showLeds(Bright0);
          LEDshow0 = false;
        } else {
          FastLED[1].showLeds(Bright1); FastLED[1].showLeds(Bright1);
          LEDshow1 = false;
        }
      }
      // reset/continue the run flags to build next patterns
      LED0_Run = true; LED1_Run = true;
      // The 20ms timer is sub-divided to develop a two-phase 40ms timer
      LED40ms = !LED40ms;               // Toggle the 40ms flag.
      if (LED40ms) {LED40msTask = 1;}   // Prime the 40ms tasks in this correct phase.
    } else {
      // 2nd priority is given to updating data for the Monitor+ app, if connected.
      // We can do this in Core0 as the overhead of creating serial data is quite low.
      if (LED40msTask > 0) {
        // These tasks are run every 40ms
        switch(LED40msTask) {
          case 1: if (DispMon) {Display_40ms();} break;
        } LED40msTask--;
      } else {
        // For Core 1 to change the task it uses the 'set' flag and 'New' value
        // Check to see if there is a request from Core 1 to change the LED tasks.
        // Core 1 can't do this directly, as Core 0 may conflict with the action.
        if (LED0_Set) {set_LED0_Task(LED0_New);}
        if (LED1_Set) {set_LED1_Task(LED1_New);}
        // One or more 'Run' tasks are performed immediately after a 'show' event.
        // Once the run tasks are complete, the flag is cleared and a 'show' flag is set
        if (LED0_Run) {GunLED_Tasks();}
        if (LED1_Run) {EyeLED_Tasks();}
      }
    }
    // Make a Beep! sound for a specified period in milliseconds.
    // Beep is set to a negative number to trigger the process.
    if (Beep != 0) {
      if (Beep < 0) {TonePlay(2000); Beep = -Beep;}
      Beep--; if (Beep < 1) {ToneDetach();}
    }

    // Return to RTOS system with a 1ms call back.
    vTaskDelay(1/portTICK_PERIOD_MS);
  }
}

// --------------------------------------------------------------------------------

void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  // Callback when data is received over Wi-Fi
  // we copy the data into a buffer and set a flag so that the main loop will act
  // to handle it
  while (WiFiRx) {
    // buffer contains unread data so deal with it before adding new
    readWiFiRx();
  }

  memcpy(&Rx_Buff, incomingData, len);

  // incoming data will always be two or more characters, ie. text == $+++
  if (len < 2) {Serial.println("Rx Error - no Rx data"); return;}
  
  if (Rx_len > 0) {Serial.println("Rx Error - buffer in use");} // this should not happen

  // if connected Rx data will have a checksum so check it first
  Rx_len = len - 1; Rx_Pnt = 0;
  RxGetChecksum(Rx_len);  // generate a checksum from Rx data
  if (Rx_Chk != Rx_Buff.ESPdata[len - 1]) {
    Serial.println("Rx Error - checksum");
  }
  // set flags before exiting
  WiFiRxBytes += Rx_len; Rx_Task = 0; WiFiRx = true;
  WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
  WiFiRxRec = true; // set flag indicating data has been recieved from Wii Transceiver
  //  Serial.print("Bytes received: "); Serial.println(len);
}

// --------------------------------------------------------------------------------

void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  // Callback when data is sent over Wi-Fi, determines success or failure
  //  t1 = micros();
  //  Serial.print("Send Status: ");
  if (WiFiConnected) {
    // the link is already active so handle normal success/failure scenarios
    if (status == 0){
  //      Serial.println("Success " + String(t1-t0));
      WiFiTx = true;    // flag WiFi Tx success
      WiFiTx_CB = 1;    // set flag to indicate that a call back state has occured
      WiFiTxErrCnt = 0; // clear the error flag in case of a glitch
      WiFiPing = 50;    // set link time-out to 2 seconds
    } else {
  //      Serial.println("Fail " + String(t1 - t0));
      WiFiTx = false;   // flag WiFi Tx failure
      WiFiTx_CB = -1;   // set flag to indicate that a resend is required
      WiFiTxErrCnt++;
    }
  } else {
    // not connected so we must have tried to connect and may get an error
    if (status == 0){
      // successful so enter the connected state
      // Serial.println("Connected!");
      WiFiConnected = true;
      WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
      WiFiTxErrCnt = 0; // reset error count for this connection
    } else {
  //      Serial.println("Status = " + String(status));
      WiFiTxErrCnt++;
    }
    WiFiTx_CB = 1;
  }
}

// --------------------------------------------------------------------------------

void resetDMs() {
  // call at the outset or whenever changing modes, as display lists are
  // different for Demo and Test modes
  DM_Control = -9;          // 'Control' display assigned Display Mode
  DM_DbBrake = -9;          // 'DeadBand & Brake' display assigned Display Mode
  DM_Drive = -9;            // 'Drive' display assigned Display Mode
  DM_Filter = -9;           // 'Filters' display assigned Display Mode
  DM_GunFire = -9;          // 'Gun Fire' display assigned Display Mode
  DM_Min = -9;              // min/max Display Mode pointer values
  DM_Max = -9;              // min/max Display Mode pointer values
  DM_Motor_Stat = -9;       // 'Motor Stats' display assigned Display Mode
  DM_Motor_Test = -9;       // 'Motor Test' display assigned Display Mode
  DM_MPU_Acc = -9;          // 'MPU Acc.' display assigned Display Mode
  DM_MPU_Gyros = -9;        // 'MPU Gyros' display assigned Display Mode
  DM_OpMode = -9;           // 'Motor Calcs' display assigned Display Mode
  DM_PtchPID = -9;          // 'Pitch PID' display assigned Display Mode
  DM_Range = -9;            // 'Range' display assigned Display Mode
  DM_SafeMode = -9;         // 'Safe Mode' display assigned Display Mode
}

// --------------------------------------------------------------------------------

void runPOST() {
  // do initialisation checks before running main code
  Serial.println(F("Running POST..."));
  Serial.print("\n\n" + Release + "  " + Date + "\n");

  // Decide whether to run in Wi-Fi mode or BLE mode, depending on whether one of the
  // button switches SW0 or SW1 is held pressed as the micro comes out of reset.
  // WiFiMode == false for Bluetooth BLE mode
  // WiFiMode == true  for normal 2.4GHz WiFi mode
  if (!digitalRead(sw0Pin) || !digitalRead(sw1Pin)) {WiFiMode = false;} else {WiFiMode = true;}

  // initilaise RGB LED arrays for both strips to OFF
  // by calling 'show' here for each strip, we can show them separately later by doing it twice
  fill_solid(Eyes_LED, 7, CRGB::Black);
  FastLED[0].showLeds(gBrightness);
  fill_solid(Guns_LED, 7, CRGB::Black);
  // colour mouth LED to match WiFiMode selections
  if (!WiFiMode) {Guns_LED[5].setRGB(  0,  0, 64);} else {Guns_LED[5].setRGB(  0, 64,  0);}
  FastLED[1].showLeds(gBrightness);
  Serial.println("RGB LEDs cleared.");

  // Set up battery reading
  initBatteryRead();
  
  // Test for MPU 6050 3-axis motion sensor and initialise
  Wire.beginTransmission(MPU_address);
  I2CErr = Wire.endTransmission();
  if (I2CErr == 0) {
    Serial.println(F("Initialising MPU6050..."));
    MPU_6050_Initialise();
  }
  if (I2CErr == 0) {I2C_MPU = true; Serial.println(F("MPU6050 Initialised"));}
  else {I2C_MPU = false; Serial.println(F("MPU6050 Not Found!"));}

  // Test for VL53L1X laser sensor and initialise
  digitalWrite(XSHUT_PIN,HIGH);  // enable VL53L1X sensor
  Wire.beginTransmission(0x29);
  I2CErr = Wire.endTransmission();
  if (I2CErr == 0) {
    I2C_LTOF = true; Serial.println("Initialising VL53L1X...");
    VL53L1X_ON();                 // turn it ON briefly to confirm that it is there
    delay(100);
    VL53L1X_OFF();                // turn the laser OFF by default
  } else {I2C_LTOF = false; Serial.println("VL53L1X Not Found!");}

  if (!USB) {
    // Battery voltage looks healthy, robot is un-restricted
  } else {
    // Battery voltage is very low, so USB assumed and restrictions apply
  }

  // POST complete
  Beep = -100;
  set_LED0_New(1);                        // set Gun LED task to intro
  Serial.println(F("POST complete"));
}

// --------------------------------------------------------------------------------

void setDefaults() {
  // load default values, declared here in case of a soft reset
  AcAvgDiv = 0;             // dividing factor used in averaging
  AccOffEn = true;          // if == false then display RAW Acc values
  A_Cnt = 0;                // counter used Test mode
  AccRampX = 0;             // max allowable rate of change of AcX, 0 == Off
  AccRampY = 0;             // max allowable rate of change of AcY, 0 == Off
  AccRampZ = 0;             // max allowable rate of change of AcZ, 0 == Off
  AcX = 0;                  // MPU raw X-acceleration value
  AcXAvg = 0;               // averaged AcX value to reduce noise
  AcXSum = 0;               // averaging accumulator
  AcY = 0;                  // MPU raw Y-acceleration value
  AcYAvg = 0;               // averaged AcX value to reduce noise
  AcYDwnCnt = 0;            // counter used in orientation auto-mode switching
  AcYUpCnt = 0;             // counter used in orientation auto-mode switching
  AcYSum = 0;               // averaging accumulator
  AcZ = 0;                  // MPU raw Z-acceleration value
  AcZAvg = 0;               // averaged AcX value to reduce noise
  AcZSum = 0;               // averaging accumulator
  Auton$ = "Autonomous";    // autonomous mode string
  AutoOFF = 0;              // auto-power OFF timer
  BackAwayTask = 0;         // back away function task pointer
  Beep = 0;                 // makes a beep sound for a specified period
  Bright0 = 255;            // FastLED 0 brightness
  Bright1 = 255;            // FastLED 1 brightness
  Brightness = 255;         // Display brightness
  C_Cnt = 0;                // counter used in 'C' button detection and Test mode
  C_Dn = 0;                 // counter for 'C' button down period
  cmdMode = ' ';            // command mode
  cmdSgn = 1;               // cmdVal sign, 0 or -1
  cmdType = ' ';            // command mode type
  cmdVal = 0;               // value associated with a cmdType
  CZ = 0;                   // received button values
  CZ_Wait = false;          // flag used for button release wait function
  DEMO = false;             // if == true then Demo Mode has been triggered
  DispBack = false;         // flag used to draw a back arrow on the display
  DispChng = false;         // flag invokes initialisation code
  DispClr = false;          // set == true to clear the memory contents
  DispCnt = 0;              // counter used in display updating
  DispDel = 0;              // >0 display delay counter, stops updates over I2C
  DispDisp = false;         // set = true to immediately display memory contents
  DispMenu = 0;             // group type, 0 = sensors, 1 = tasks, etc
  DispMode = 1;             // the type of display for a given mode
  DispModeCnt = 0;          // counter used in display updating
  DispMon = false;          // == true if connected to OLED Mirror app
  DispNext = 0;             // used to transition to another DispMode
  DispNOP = false;          // set == true to stop display I2C updates
  DispNow = 0;              // if > 0 display special modes
  DispNxtCnt = 0;           // counter used in auto-branching of display
  DispPnt = 0;              // points at the current display mode, default = 0 batteries
  DispTask = 0;             // task pointer for display functions
  DispTx = 0;               // counter used with display mirror app
  DispVal = 0;              // general display value, passed from tasks
  EnStepLft = false;        // if == true then enable left stepper clock
  EnStepRht = false;        // if == true then enable right stepper clock
  Gear = 1;                 // speed control flag, default = 0
  GPmax = 0;                // end of graphing list, for Scope mode
  GPmin = 0;                // start of graphing list, for Scope mode
  GPnn = 0;                 // pointer to current graphing items, default = 0
  GP_Title = false;         // if == true then report a title in Scope graphing mode
  GunFire = 0;              // > 0 when firing the guns
  GyAvgDiv = 0;             // dividing factor used in averaging
  GyCal = 250;              // counter used to determine gyro calibration period
  GyrOffEn = true;          // if == false then display RAW Gyro values
  GyX = 0;                  // MPU raw X-gyroscope value
  GyXAvg = 0;               // averaged GyX value to reduce noise
  GyXOff = 0;               // offset to be added to raw X-gyroscope value
  GyXOffCnt = 0;            // offset counter for gyro centering
  GyXSum = 0;               // averaging accumulator
  GyY = 0;                  // MPU raw Y-gyroscope value
  GyYAvg = 0;               // averaged GyY value to reduce noise
  GyYOff = 0;               // offset to be added to raw Y-gyroscope value
  GyYOffCnt = 0;            // offset counter for gyro centering
  GyYSum = 0;               // averaging accumulator
  GyZ = 0;                  // MPU raw Z-gyroscope value
  GyZAvg = 0;               // averaged GyZ value to reduce noise
  GyZOff = 0;               // offset to be added to raw Z-gyroscope value
  GyZOffCnt = 0;            // offset counter for gyro centering
  GyZSum = 0;               // averaging accumulator
  JoyX = 128;               // received value, default
  JoyY = 128;               // received value, default
  LED40msTask = 0;          // 40ms tasks running on Core 0
  LED0_Del = 0;             // gun LED Task delay, adds pauses into the processes
  LED1_Del = 0;             // eye LED Task delay, adds pauses into the processes
  LED0_Flash = 0;           // if > 0 then flash Gun LEDs bright white for n counts
  LED1_Flash = 0;           // if > 0 then flash Eye LEDs bright white for n counts
  LED0_Last = 0;            // record previous LED0 task
  LED1_Last = 0;            // record previous LED1 task
  LED0_New = 0;             // pointer to new Gun LED task
  LED1_New = 0;             // pointer to new Eye LED task
  LED0_Run = true;          // == true if LED_Task is to run
  LED1_Run = true;          // == true if LED_Task is to run
  LED0_Set = false;         // flag used to initiate Gun LED task change
  LED1_Set = false;         // flag used to initiate Eye LED task change
  LEDshow0 = false;         // LED show flag for face LEDs
  LEDshow1 = false;         // LED show flag for gun LEDs
  LED0_Task = -1;           // Gun LED task pointer
  LED1_Task = -1;           // Eye LED task pointer
  LED_Tasks = 0;            // LED 20ms task pointer
  Lft_Dir = LOW;            // left motor DIR pin status
  Lft_En = HIGH;            // left motor enable, active LOW, default = HIGH
  Lft_Motor_Cnt = 0;        // left motor counter
  Lft_Step_Cnt = 0;         // left motor step counter
  Lft_Motor_Tgt = 0;        // left motor target count
  Lft_Motor_Throttle = 1000;// left motor throttle demand
  Lft_Step_Inc = 1;         // left motor step increment
  LTOFcalc = false;         // if == true then calculate speed of ranged object in m/s
  LTOFspd10 = 0;            // speed of ranged target in m/s x 10
  LTOFspd10L = 0;           // previous speed of ranged target in m/s x 10
  LTOFspdCnt = 0;           // speed peak detector sustain limit
  M210 = 0b010;             // right driver m pin assignments
  MainMode = 1;             // determined after RESET, based on orientation, 0 = Demo, 1 = Test
  MainType = 0;             // a number representing the type of task, default = 0
  Motor_Test = false;       //  if == true then conduct motor tests, default = false
  moveDir = 0;              // move direction flag. 0 == stationary
  MyStatus = "Waking Up";   // status displayed in MOnitor+ app
  NoteLast = NOTE_C;        // most recently played note, default = middle 'C'
  Octave = 4;               // most recently played octave, default
  OpMode = 0;               // determines the output calculation method, 0, 1,etc
  Ping = 0;                 // 'ping' counter
  PrintDel = 0;             // delay counter used in serial printing
  PrintTgt = 0;             // 0 == serial port, 1 == WiFi ESP-NOW
  PrintTx = "";             // printed strings
  Ptch_d_En = true;         // PID d-gain enable true=ON, false=OFF
  Ptchen = false;           // if == true then Pitch PID mode is enabled
  PtchGo = false;           // == true when Pitch axis is within starting angle
  Ptch_i_En = true;         // PID i-gain enable true=ON, false=OFF
  PtchStepT = 0;            // Pitch impulse step test duration in x5ms (default = 0)
  Pwr = 0.5;                // PID output power setting, 0.5 to 1.0
  RangeCntMax = 10;         // discriminator counter, adjustable limit
  RangeEn = false;          // == true if VL53L1X ranging is active
  RangeFps = true;          // == true if LTOF fps is to be displayed
  RangeMax = LtofLimit;     // the current limited to Range measurements
  RangeMin = LtofMin;       // the current lower limited to Range measurements
  RangeNew = false;         // set == true when a new range is available
  RangePkPk =  8;           // Range pk-pk noise limit (default 20)
  RangeRate = 64;           // max rate of inter-measurement Range change, default = 64
  RangeRawF = 1000;         // rate limited RangeRaw value (default)
  Ready = false;            // if == true then we have completed setup tasks, default = false
  readWiiCall = false;      // flag prevents call stack overflow
  Report20ms = 0;           // timer used in serial port reporting, for debug
  Rht_Dir = LOW;            // right motor DIR pin status
  Rht_En = HIGH;            // right motor enable, active LOW, default = HIGH
  Rht_Motor_Cnt = 0;        // right motor counter
  Rht_Step_Cnt = 0;         // right motor step counter
  Rht_Motor_Tgt = 0;        // right motor target count
  Rht_Motor_Throttle = 1000;// right motor throttle demand
  Rht_Step_Inc = -1;        // right motor step increment
  RxByte = 0;               // byte containing most recent WiFi control data
  RxClass = false;          // set == true when a Wii Classic command has been receieved
  Rx_Chk = 0;               // Rx checksum
  RxCZ = 0;                 // buffered received CZ value
  RxJoyX = 0;               // buffered received JoyX value
  RxJoyY = 0;               // buffered received JoyY value
  Rx_len = 0;               // length of WiFi received data block
  RxNunch = false;          // set == true when a Wii Nunchuk command has been receieved
  RxRec = false ;           // set true when a valid frame of data is received
  RxState = 0;              // receiver state machine state
  Rx_Task = 0;              // pointer user by Rx state machine
  RTxTimeout = 0;           // counter used to auto-reset if WiFi fails
  RxVal = -1;               // value received from serial Rx
  SafeMode = -1;            // controls the state of the control system
  Sb = true;                // if == true then perform self-balance setpoint adjustments (true)
  SerialRx = false;         // == true if any character has been received
  StepLft = LOW;            // left stepper driver step clock
  StepLftHi = LOW;          // == HIGH if left stepper driver step clock is HIGH
  StepLftPeriod = 5000;     // left stepper clock period in µs (5000 = 200Hz, 15rpm)
  StepRht = LOW;            // right stepper driver step clock
  StepRhtHi = LOW;          // == HIGH if right stepper driver step clock is HIGH
  StepRhtPeriod = 5000;     // right stepper clock period in µs (5000 = 200Hz, 15rpm)
  sw0Cnt = 0;               // button switch counter
  sw0DwnTime = 0;           // button pressed down time
  sw0LastState = HIGH;      // previous state of button switch, HIGH/LOW
  sw0_Nop = false;          // if == true don't perform switch functions
  sw0State = HIGH;          // state of read button switch pin
  sw0Timer = 0;             // timer used to detemine button sequences
  sw0Wup = false;           // set == true for SW0 to wait for button release
  sw1Cnt = 0;               // button switch counter
  sw1LastState = HIGH;      // previous state of button switch, HIGH/LOW
  sw1_Nop = false;          // if == true don't perform switch functions
  sw1State = HIGH;          // state of read button switch pin
  sw1Timer = 0;             // timer used to detemine button sequences
  sw1Wup = false;           // set == true for SW1 to wait for button release
  Task4ms = 0;              // 4ms task pointer
  Task8ms = 0;              // 8ms task pointer
  Task20ms = 0;             // 20ms task pointer
  Task40ms = 0;             // 40ms task pointer
  Turning = 0;              // value used to track applied controller demands
  TurnSpd = 16;             // value applied to right/left when turning
  ToneAtt = false;          // if == true then tone channel is attached, default = false
  ToneDel = 0;              // tone delay counter
  ToneFreq = 0;             // frequency used in TonyPlay() function
  TonePnt = 0;              // tone task pointer
  VL53L1X_OC = 199;         // VL53L1X ROI SPAD centre
  VL53L1X_ROIX = 16;        // VL53L1X ROI SPAD width
  VL53L1X_ROIY = 16;        // VL53L1X ROI SPAD height
  VL53L1X_SkipDel = 33;     // read function skip timer delay tracker, nominally 33
  VL53L1X_Task = 0;         // if > 0 then run LV53L1X tasks
  WiFiCntC = 0;             // C button counter for WiFi enable
  WiFiConCnt = 0;           // counter used in connection retry times
  WiFiConnected = false;    // == true if a WiFi connection has been established
  WiFiEn = true;            // == true to enable WiFi control
  WiFiPing = 0;             // Rx counter used to maintain connected state
  WiFiRx = false;           // true when a block of data has been received
  WiFiRxBytes = 0;          // number of bytes receiv ed over the WiFi link
  WiFiRxErr = 0;            // number of receive errors in the WiFi link
  WiFiRxRec = false;        // true when a block of data has been successfully received
  WiFiTryCnt = 0;           // WiFi try to connect count
  WiFiTryNum = 0;           // WiFi try to connect total count
  WiFiTx = false;           // true when a block of data has been sent successfully
  WiFiTxBytes = 0;          // number of bytes receiv ed over the WiFi link
  WiFiTx_CB = 1;            // == 1 call back is clear to send, == -1 resend
  WiFiTx_Chk = 0;           // Tx checksum
  WiFiTx_len = 0;           // >0 when WiFi Tx buffer contains data to be sent
  WiFiTxErr = 0;            // WiFi data error
  WiFiTxErrCnt = 0;         // WiFi Tx error count, leads to disconnection
  WiiBYAX = 0b00001111;     // Wii Classic right-hand XYAB buttons
  WiiDigi = 0b00001111;     // Wii Classic digital pad bits
  WiiSHS  = 0b00000111;     // Wii Classic centre select,home,start bits
  WiiType = 0;              // device type, 0 == Nunchuk, 1 == Classic
  Z_Dn = 0;                 // Z-button down count, 0 == released, default

  Set_Speeds(1);            // set the speed/power settings
}

// --------------------------------------------------------------------------------

void setPIDgains() {
  // Called after reset, to set the PID gains to their default values.
  // You determine the optimum values for each variable using the Monitor+
  // then include them here as firmware.
  LpFc = 0.900;           // low pass filter co-efficient, between 1.00 to 0.01 (min/max)
  Ptch_Brake = 0.015;     // brake factor applied to PID output
  Ptch_Db = 5.0;          // Pitch deadband limit in terms of Pitch_Op
  Ptch_p_gain = 20.0;     // Gain setting for Pitch P-controller (1.0)
  Ptch_i_gain = 1.50;     // Gain setting for Pitch I-controller (0.0)
  Ptch_i_max = 300.0;     // limit Pitch I-controller output(2.0)
  Ptch_d_gain = 4.0;      // Gain setting for Pitch D-controller (0.0)
  Ptch_OpBrake = 10.0;    // PID output at which the brake function is applied
  Ptch_out_max = 300.0;   // limit Pitch PID output to limit throttle min to 10
  PtchStartAng = 0.30;    // initial target angle for balancing Pitch (default 0.0)
  SbSpInc = 0.0015;       // Pitch self-balance auto-increment 0.0015
}

// --------------------------------------------------------------------------------

void synchLoopTimers() {
  // reset main loop timers
  while (millis() < 40){}     // we must wait until millis() is at least 40 from reset
  DispBatWarn = millis();     // 30 sec timer used to prevent battery warnings being too frequent
  DispOverun = millis();      // timer used to prevent the display function being over-run
  LED20ms = millis() + 20;    // set LED 20ms timer
  LED40ms = false;            // a 40ms timer derived from the LED20ms event
  next4ms = millis() - 3;    // set loop timer
  print40ms  = millis() - 12; // reset 40ms printer timer with 6ms phase shift
  //  Serial.print("synchLoopTimers()");
}

// --------------------------------------------------------------------------------

